/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/
/*! \file DAG.h
\brief API of the DAG class.

*/

#include "DAG.h"
#include "loop.h"
#include <LEDA/graph/templates/shortest_path.h>
#include "LEDA/graph/shortest_path.h" /*belman ford, shortest paths */

#include <xercesc/dom/DOM.hpp>
#include <xercesc/dom/DOMDocument.hpp>
#include <xercesc/dom/DOMDocumentType.hpp>
#include <xercesc/dom/DOMElement.hpp>
#include <xercesc/dom/DOMImplementation.hpp>
#include <xercesc/dom/DOMImplementationLS.hpp>
#include <xercesc/dom/DOMNodeIterator.hpp>
#include <xercesc/dom/DOMNodeList.hpp>
#include <xercesc/dom/DOMText.hpp>

#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/util/XMLUni.hpp>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

XERCES_CPP_NAMESPACE_USE

using namespace LEDA;
using namespace std;

/* some useful function */
namespace DDG{











/*------------------ DAG -----------------------------------*/
DAG::~DAG()
{
}


DAG::DAG()
	: down(), up(),
  longuest_path_from(), longuest_path_to(), shortest_path_from(), shortest_path_to(), V_R(), lp_array(), E_R(), Cons(), ISA(), node_index_table(nil, MAX_NODE)
{
  
  ISA.set_default_arch();
  cpt_nodes=0; //initialisation of the counter of nodes

  /*flag initialisation */
  
}	    

DAG::DAG(const ARCHITECTURE isa)
	: down(), up(),
	longuest_path_from(), longuest_path_to(), shortest_path_from(), shortest_path_to(), V_R(), lp_array(), E_R(), Cons(), node_index_table(nil, MAX_NODE)
{
	
	cpt_nodes=0; //initialisation of the counter of nodes
	/*flag initialisation */
  	 // Read the instructions types
  	 ISA=isa;
}	
    
void DAG::clear(){
  	int i; 
	GRAPH<INFO_NODE, INFO_EDGE>::clear();
	cpt_nodes=0; //initialisation of the counter of nodes
	node_index_table.clear();
	Cons.clear();
	Sources.clear();
	Targets.clear();
	up=node_array<set<node> > ();
	shortest_path_from= node_array<int> ();
	shortest_path_to=  node_array<int> ();
	lp_array=node_matrix<int>();
	down=node_array<set<node> > ();
	E_R.clear();
	V_R.clear();
	
}

int DAG::new_node(int it_id)
{
	INSTRUCTIONS_TYPES opt = ISA.get_instruction_type_by_id(it_id);
	REGISTER_TYPES rt;
	cpt_nodes++;
	node n;
  	INFO_NODE in(cpt_nodes, opt);
  	n=GRAPH<INFO_NODE, INFO_EDGE>::new_node(in);
  	node_index_table[cpt_nodes]=n;
  	leda::list<int> lr=opt.get_list_regtypes_id();
  	int t;
  	forall(t, lr){
  		 	V_R[t].insert(n);
  	}
  	
  	
  	return cpt_nodes;
}

int DAG::new_node(int it_id, int n_id)
{
	INSTRUCTIONS_TYPES opt = ISA.get_instruction_type_by_id(it_id);
	node n;
  	INFO_NODE in(n_id, opt);
  	if (node_index_table[n_id]!=nil) throw NonUniqueNodeID(n_id);
  	
  	n=GRAPH<INFO_NODE, INFO_EDGE>::new_node(in);
  	node_index_table[n_id]=n;
  	leda::list<int> lr=opt.get_list_regtypes_id();
  	int t;
  	forall(t, lr){
  	 	V_R[t].insert(n);
  	 }
  	if(n_id> cpt_nodes )  cpt_nodes = n_id; 
  	return n_id;
}



int DAG::new_node(int it_id, int n_id, std::string text_inst)
{
	INSTRUCTIONS_TYPES opt = ISA.get_instruction_type_by_id(it_id);
	node n;
	INFO_NODE in(n_id, opt, text_inst);
	if (node_index_table[n_id]!=nil) throw NonUniqueNodeID(n_id);
	
	n=GRAPH<INFO_NODE, INFO_EDGE>::new_node(in);
	node_index_table[n_id]=n;
  	leda::list<int> lr=opt.get_list_regtypes_id();
  	int t;
  	forall(t, lr){
  		V_R[t].insert(n);
  	 }
  	
  	if(n_id> cpt_nodes )  cpt_nodes = n_id;
	return n_id;
}

int DAG::new_node(int it_id, std::string text_inst)
{
	INSTRUCTIONS_TYPES opt = ISA.get_instruction_type_by_id(it_id);
	cpt_nodes++;
	node n;
	INFO_NODE in(cpt_nodes, opt, text_inst);
	n=GRAPH<INFO_NODE, INFO_EDGE>::new_node(in);
	node_index_table[cpt_nodes]=n;
  	leda::list<int> lr=opt.get_list_regtypes_id();
  	int t;
  	forall(t, lr){
  	 	V_R[t].insert(n);
  	 }
  	return cpt_nodes;
}


	
	
void DDG::DAG::del_node(node n)
{
	INSTRUCTIONS_TYPES ins=instruction_type(n);
	int t;
	leda::list<int> lt=ISA.get_list_regtypes_id();
	
	if((instruction_type(n)).is_value()){
		leda::list<int> lr=ins.get_list_regtypes_id();
		forall(t, lr){			
			V_R[t].del(n);
		}
	}
	forall(t, lt){
		set<edge>ER=E_R[t]; // to remove the set of flow edges through registers incident to n
		
		ER &= ( list_to_set(in_edges(n))
			+ // union
			list_to_set(out_edges(n))
			) ;// now ER contains all flow edges through registers incident to n
	
		E_R[t] -=  ER; // now, we remove the flow edges 	through registers incident to n
	}
	node_index_table[id(n)]=nil;	
	GRAPH<INFO_NODE, INFO_EDGE>::del_node(n);
}

void DDG::DAG::hide_node(node n)
{
	INSTRUCTIONS_TYPES ins=instruction_type(n);
	leda::list<int> lt=ISA.get_list_regtypes_id();
	int t;

	if((instruction_type(n)).is_value()){
		leda::list<int> lr=ins.get_list_regtypes_id();
		forall(t, lr){
			V_R[t].del(n);

		}
	}
	forall(t, lt){
		set<edge>ER=E_R[t]; // to remove the set of flow edges through registers incident to n
		
		ER &= ( list_to_set(in_edges(n))
			+ // union
			list_to_set(out_edges(n))
			) ;// now ER contains all flow edges through registers incident to n
	
		E_R[t] -=  ER; // now, we remove the flow edges 	through registers incident to n
	}
	node_index_table[id(n)]=nil;
	GRAPH<INFO_NODE, INFO_EDGE>::hide_node(n);
}

void DDG::DAG::hide_node(node n, LEDA::list<edge>& h_edges) {
	INSTRUCTIONS_TYPES ins=instruction_type(n);
	leda::list<int> lt=ISA.get_list_regtypes_id();
	int t;
	
	if((instruction_type(n)).is_value()){
		leda::list<int> lr=ins.get_list_regtypes_id();
		int t;
		forall(t, lr){
			V_R[t].del(n);
		}
	}
	
	forall(t, lt){
	
		
		set<edge>ER=E_R[t]; // to remove the set of flow edges through registers incident to n
		
		ER &= ( list_to_set(in_edges(n))
			+ // union
			list_to_set(out_edges(n))
			) ;// now ER contains all flow edges through registers incident to n
	
		E_R[t] -=  ER; // now, we remove the flow edges 	through registers incident to n
	}
	node_index_table[id(n)]=nil;
	GRAPH<INFO_NODE, INFO_EDGE>::hide_node(n, h_edges);
}

void DDG::DAG::restore_node(node u)
{
	int t;
	INSTRUCTIONS_TYPES ins;
	set<edge> ER;
	edge e;
	GRAPH<INFO_NODE, INFO_EDGE>::restore_node(u);
	node_index_table[id(u)]=u;
	ins=instruction_type(u);
	if(ins.is_value()){
		leda::list<int> lr=ins.get_list_regtypes_id();
		int t;
		forall(t, lr){
			V_R[t].insert(u);
		}
		
	}
	ER=list_to_set(in_edges(u))+ list_to_set(out_edges(u));
	forall(e,ER){
		INFO_EDGE ie=inf(e);
		if(ie.is_flow_reg()) {
			t=get_register_type_id(e);
			if (t<0){
				cerr<<"DDG warning :flow dependance through register with register type -1.\n";
			}
			else E_R[t].insert(e);
		}
	}
	
}

void DDG::DAG::restore_all_nodes()
{
	node u;
	int t;
	INSTRUCTIONS_TYPES ins;
	set<edge> ER;
	edge e;
	GRAPH<INFO_NODE, INFO_EDGE>::restore_all_nodes();
	
	
	forall_nodes(u, *this){
		node_index_table[id(u)]=u;
		ins=instruction_type(u);
		if(ins.is_value()){
		leda::list<int> lr=ins.get_list_regtypes_id();
		int t;
		forall(t, lr){
			V_R[t].insert(u);
		}
		
		}
		ER=list_to_set(in_edges(u))+ list_to_set(out_edges(u));
		forall(e,ER){
		INFO_EDGE ie=inf(e);
		if(ie.is_flow_reg()) {
			t=get_register_type_id(e);
			if (t<0){
				cerr<<"DDG warning :flow dependance through register with register type -1.\n";
			}
			else E_R[t].insert(e);
		}
		}
	}
	
	
}

void DDG::DAG::del_node(int i)
{
	node n;
	try{
		n=node_with_id(i);
	}
	catch(IllegalAccessNode)
	{
		return ;
	}
	del_node(n);
}



void DDG::DAG::del_edge(edge e)
{
	INFO_EDGE ie=inf(e);
	if(ie.is_flow_reg()) {
		int t=get_register_type_id(e);
		if (t<0){
			cerr<<"DDG warning :flow dependance through register with register type -1.\n";
		}
		else E_R[t].del(e);
	}
	
	GRAPH<INFO_NODE, INFO_EDGE>::del_edge(e);
}

void DDG::DAG::hide_edge(edge e)
{
	INFO_EDGE ie=inf(e);
	if(ie.is_flow_reg()) {
		int t=get_register_type_id(e);
		if (t<0){
			cerr<<"DDG warning :flow dependance through register with register type -1.\n";
		}
		else E_R[t].del(e);
	}
	GRAPH<INFO_NODE, INFO_EDGE>::hide_edge(e);
}

void DDG::DAG::restore_edge(edge e){
	GRAPH<INFO_NODE, INFO_EDGE>::restore_edge(e);
	INFO_EDGE ie=inf(e);
	if (ie.is_flow_reg()){
		int t=get_register_type_id(e);
		if (t<0){
			cerr<<"DDG warning :flow dependance through register with register type -1.\n";
		}
		else E_R[t].insert(e);
	}
}

void DDG::DAG::restore_all_edges()
{
	edge e;
	GRAPH<INFO_NODE, INFO_EDGE>::restore_all_edges();
	forall_edges(e, *this) {
		INFO_EDGE ie=inf(e);
		if (ie.is_flow_reg()){
			int t=get_register_type_id(e);
			if (t<0){
				cerr<<"DDG warning :flow dependance through register with register type -1.\n";
			}
		}
		
	}
}

void DDG::DAG::del_all_edges(){
	E_R.clear();
	GRAPH<INFO_NODE, INFO_EDGE>::del_all_edges();
}

edge DAG::new_edge(int s, int t, INFO_EDGE ie)
{
  node sourc,targe;
  edge e;
  try{
    sourc=node_with_id(s);
    targe=node_with_id(t);
  }
  catch(IllegalAccessNode)
    {
      return NULL;
    }
   e=GRAPH<INFO_NODE, INFO_EDGE>::new_edge(sourc, targe, ie);
   if (ie.is_flow_reg()){
     int t=ie.get_register_type_id();
     if (t<0) cerr<<"DDG Warning : a flow edge through register with a register type equal to -1.\n";
     else E_R[t].insert(e);
   }
  return e;
}

edge DAG::new_edge(node sourc, node targe, INFO_EDGE ie)
{
  edge e;
  e=GRAPH<INFO_NODE, INFO_EDGE>::new_edge(sourc, targe, ie);
  if (ie.is_flow_reg()){
     int t=ie.get_register_type_id();
     if (t<0) cerr<<"DDG Warning : a flow edge through register with a register type equal to -1.\n";
     else E_R[t].insert(e);
   }
  return e;
}



/* all the edges from u to v */
set<edge> DAG::edges_from_to(node u, node v) const {
  edge e;
  set<edge> out_edge_u=list_to_set(out_edges(u));
  forall(e, out_edge_u){
    if(target(e)!=v) out_edge_u.del(e);
  }
  return out_edge_u;
}

set<edge> DAG::edges_from_to(int idu, int idv) const {
  edge e;
  node u, v;
  set<edge> out_edge_u;
  try{
    u=node_with_id(idu);
    v=node_with_id(idv);
  }
  catch(IllegalAccessNode)
    {
      return out_edge_u;
    }
  return edges_from_to(u, v);
}

node DAG::node_with_id(int i) const
{
	if (node_index_table[i]==nil)
		throw IllegalAccessNode(i);
	else {
		if (is_hidden(node_index_table[i])) {
			cerr<<"DDG error. The node with id "<<i<<" is hidden.\n";
			throw IllegalAccessNode(i);
		}
		return node_index_table[i];
	}
}

node DAG::node_with_name(std::string name) const
{
  node n;
  forall(n, node_index_table){
    if(get_string_attribute(n) == name){
      if(is_hidden(n)){
        cerr<<"DDG error. The node with name "<< name <<" is hidden.\n";
        throw IllegalAccessNode(0);
      }
      return n;
    }
  }
  throw IllegalAccessNode(0);
}

/* transitive closure, etc.. */
void DAG::build_internal_structures(){
	GRAPH <node,edge>  G_c;
	node u,v, n;
	edge e;
	int t;
	INFO_EDGE ie;
	INFO_NODE in;
	set<node> s, s1, s2;
	list<edge> le;
	leda::list<int> lr;
	edge_array<int> cost(*this);
	node_matrix<int> all_pair_longuest_path(*this), all_pair_shortest_path(*this);
	
	down.init(*this);
	up.init(*this);
	/* set of values in G */
	V_R.clear();
	forall_nodes(u, *this){
		if ((instruction_type(u)).is_value()) {
			lr=(instruction_type(u)).get_list_regtypes_id();
			forall(t, lr){
				V_R[t].insert(u);
			}
			
		}
	}
	/*------------- TRANSITIVE CLOSURE ------------*/
	G_c=TRANSITIVE_CLOSURE(*this);
	forall_edges(e, G_c){
	u=G_c.source(e);
	u=G_c.inf(u);
	in=inf(u);
	u=node_with_id(in.id());
	v=G_c.target(e);
	v=G_c.inf(v);
	in=inf(v);
	v=node_with_id(in.id());
	up[v].insert(u);
	down[u].insert(v);
	}
	/*---------------Consumers ----------------------------*/
	
	Cons.clear();
	lr=ISA.get_list_regtypes_id();
	forall(t, lr){
		forall(e, E_R[t]){
			Cons(source(e),t).insert(target(e));
		}
	}
	/*---------- sources and sinks ---------------------*/
	Sources.clear();
	Targets.clear();
	forall_nodes(u, *this){
	if(indeg(u)==0) Sources.insert(u);
	if(outdeg(u)==0) Targets.insert(u);
	
	}
	
	/*----------- LONGUEST PATH FROM/TO NODES -----------*/
	longuest_path_from.init(*this);
	longuest_path_to.init(*this);
	lp_array.init(*this);
	/* we give a negative cost to each arc to compute the longuest path
	instead of shortest one */
	forall_edges(e, *this){
	cost[e]=-delta(e);
	}
	ALL_PAIRS_SHORTEST_PATHS_T(*this, cost, all_pair_longuest_path);
	forall(v, Sources){
	longuest_path_to[v]=0;
	}
	forall(v, Targets){
	longuest_path_from[v]=0;
	}
	
	forall_nodes(u, *this){
	if(!Sources.member(u))
	forall(v, Sources){
	if((longuest_path_to[u]< -all_pair_longuest_path(v,u)) && (down[v].member(u))){
		longuest_path_to[u]= -all_pair_longuest_path(v,u);
	}
	}
	if( !Targets.member(u))
	forall(v, Targets){
	if((longuest_path_from[u]< -all_pair_longuest_path(u,v)) &&(up[v].member(u)))
		longuest_path_from[u]= -all_pair_longuest_path(u,v);
	}
	
	}
	forall_nodes(u, *this){
	forall_nodes(v, *this){
	lp_array(u,v)=-all_pair_longuest_path(u,v);
	}
	}
	
	/*----------- SHORTEST PATH FROM/TO NODES -----------*/
	shortest_path_from.init(*this);
	shortest_path_to.init(*this);
	/* we give a the layency as the cost to each arc to compute the shortest path */
	forall_edges(e, *this){
	cost[e]=delta(e);
	}
	ALL_PAIRS_SHORTEST_PATHS_T(*this, cost, all_pair_shortest_path);
	forall(v, Sources){
	shortest_path_to[v]=0;
	}
	forall(v, Targets){
	shortest_path_from[v]=0;
	}
	
	forall_nodes(u, *this){
		if(!Sources.member(u))
		forall(v, Sources){
		if((shortest_path_to[u] < all_pair_shortest_path(v,u))&&(down[v].member(u))){
			shortest_path_to[u]= all_pair_shortest_path(v,u);
		}
		}
		if( !Targets.member(u))
		forall(v, Targets){
			if((shortest_path_from[u]<all_pair_shortest_path(u,v))&&(up[v].member(u)))
			shortest_path_from[u]= all_pair_shortest_path(u,v);
		}
		
	}
	// --check edges latencies vs operations latencies
	// 	forall_edges(e, *this)
	// 	{
	// 		if (latency(source(e)) > lp(source(e), target(e))){
	// 			it=instruction_type(source(e));
	// 		cerr<<"DDG warning : the latency of the operation "<<id(source(e))<<" of type "<<it.opcode()<<" is larger than the longest path between the nodes "<<id(source(e))<<" and "<<id(target(e))<<endl;
	// 		}
	// 	}
}


void DAG::write_to_vcg(const char *filename) const
{
  
		ofstream ofile(filename);
	node v, src, dest;
	edge e;
	INSTRUCTIONS_TYPES it;
	INFO_EDGE info_edge;
	INFO_NODE in;
	LEDA::set<node> SN;
	REGISTER_TYPES t;
	LEDA::list<REGISTER_TYPES> T=ISA.T();
	 
	if(!ofile){
	cerr<< "DDG IO failure: could not create "<<filename<<'\n';
		return;
	}
	
	
	ofile<<"graph: {\ntitle: \""<<filename<<"\"\nnode.shape : ellipse\ninfoname1 : \"ISA opcode\"\nsplines : yes\nedge.color : black\nlayout_downfactor:2\nlayout_upfactor:1\nlayoutalgorithm : maxdepth\ndisplay_edge_labels : yes\n";
	ofile<<"infoname2 : \"Latencies\"\n";
	ofile<<"infoname3 : \"Textual Attribute\"\n";
	
	ofile<<"classname "<< flowdep_reg<<" : \"" <<edge_type_to_string(flowdep_reg)<< " \"\n";
	ofile<<"classname "<< antidep_reg<<" : \"" <<edge_type_to_string(antidep_reg)<< " \"\n";
	ofile<<"classname "<< outputdep_reg<<" : \"" <<edge_type_to_string(outputdep_reg)<< " \"\n";
	ofile<<"classname "<< inputdep_reg<<" : \"" <<edge_type_to_string(inputdep_reg)<< " \"\n";
	ofile<<"classname "<< flowdep_mem<<" : \"" <<edge_type_to_string(flowdep_mem)<< " \"\n";
	
	ofile<<"classname "<< antidep_mem<<" : \"" <<edge_type_to_string(antidep_mem)<< " \"\n";
	ofile<<"classname "<< outputdep_mem<<" : \"" <<edge_type_to_string(outputdep_mem)<< " \"\n";
	ofile<<"classname "<< inputdep_mem<<" : \"" <<edge_type_to_string(inputdep_mem)<< " \"\n";
	ofile<<"classname "<< spilldep_mem<<" : \"" <<edge_type_to_string(spilldep_mem)<< " \"\n";
	ofile<<"classname "<< other_mem<<" : \"" <<edge_type_to_string(other_mem)<< " \"\n";
	
	ofile<<"classname "<< serial<<" : \"" <<edge_type_to_string(serial)<< " \"\n";
	ofile<<"classname "<< killerdep<<" : \"" <<edge_type_to_string(killerdep)<< " \"\n";
	ofile<<"classname "<< reusedep<<" : \"" <<edge_type_to_string(reusedep)<< " \"\n";
	
	forall_nodes(v, *this){
		it=instruction_type(v);
		bool isvalue=false;
		
		ofile<<"node: { title: \""<<id(v)<< "\"  info1: \""<<it.opcode()<<"\" info2: \" writes " ;
		forall(t, T){
			 if (this->is_value(v,t.get_register_type_id())){
			 	isvalue=true;
			 	ofile<<t.get_register_typename()<<"("<<t.get_register_type_id()<<") ";
		 }	
		}
		if(! isvalue){
			ofile <<". lat="<<latency(v)<<", dr="<<delta_r(v);
		}
		ofile <<"\" info3: \""<<get_string_attribute(v)<< "\" color:";
		if(isvalue){
			ofile<<"green}\n";
		}
		else ofile<<"lightgrey}\n";
		
	}
  // unparsing edges in vcg format
	forall_edges(e, *this){
		src = source(e);
		dest = target(e);
		INFO_EDGE ie=inf(e);
		ofile<<"edge: { sourcename: \"";
		ofile<<id(src);
		
		
		ofile<<"\" targetname: \"";
		
		
		ofile<<id(dest);
		
		if (ie.get_register_type_id()<0)
			ofile<<"\" class:"<<etype(e)<<" label:\""<< delta(e)<<"\" color:";
		else
			ofile<<"\" class:"<<etype(e)<<" label:\""<< delta(e)<< "," << this->ISA.get_register_typename(ie.get_register_type_id()) <<"\" color:";
		switch (etype(e)){
		case flowdep_reg: 
		case flowdep_mem: ofile<< "red}\n";
			break;		
		case antidep_reg: 
		case antidep_mem: ofile<<"green}\n";
			break;
		case outputdep_reg: 
		case outputdep_mem: ofile<<"darkmagenta}\n";
			break;	
		case inputdep_reg:
		 	inputdep_mem :ofile<<"blue}\n";
			break;
		default: serial: ofile<<"black}\n";
		}		
	}// end for edge
	ofile<<"}\n"; 
	
}

void DAG::write_to_vcg(const char *filename, int t) const
{
  
	ofstream ofile(filename);
	node v, src, dest;
	edge e;
	INSTRUCTIONS_TYPES it;
	INFO_EDGE info_edge;
	INFO_NODE in;
	LEDA::set<node> SN;
		 
	if(!ofile){
		cerr<< "DDG IO failure: could not create "<<filename<<'\n';
		return;
	}
	
	
	ofile<<"graph: {\ntitle: \""<<filename<<"\"\nnode.shape : ellipse\ninfoname1 : \"ISA opcode\"\nsplines : yes\nedge.color : black\nlayout_downfactor:2\nlayout_upfactor:1\nlayoutalgorithm : maxdepth\ndisplay_edge_labels : yes\n";
	ofile<<"infoname2 : \"Latencies\"\n";
	ofile<<"infoname3 : \"Textual Attribute\"\n";
	
	ofile<<"classname "<< flowdep_reg<<" : \"" <<edge_type_to_string(flowdep_reg)<< " \"\n";
	ofile<<"classname "<< antidep_reg<<" : \"" <<edge_type_to_string(antidep_reg)<< " \"\n";
	ofile<<"classname "<< outputdep_reg<<" : \"" <<edge_type_to_string(outputdep_reg)<< " \"\n";
	ofile<<"classname "<< inputdep_reg<<" : \"" <<edge_type_to_string(inputdep_reg)<< " \"\n";
	ofile<<"classname "<< flowdep_mem<<" : \"" <<edge_type_to_string(flowdep_mem)<< " \"\n";
	
	ofile<<"classname "<< antidep_mem<<" : \"" <<edge_type_to_string(antidep_mem)<< " \"\n";
	ofile<<"classname "<< outputdep_mem<<" : \"" <<edge_type_to_string(outputdep_mem)<< " \"\n";
	ofile<<"classname "<< inputdep_mem<<" : \"" <<edge_type_to_string(inputdep_mem)<< " \"\n";
	ofile<<"classname "<< spilldep_mem<<" : \"" <<edge_type_to_string(spilldep_mem)<< " \"\n";
	ofile<<"classname "<< other_mem<<" : \"" <<edge_type_to_string(other_mem)<< " \"\n";
	
	ofile<<"classname "<< serial<<" : \"" <<edge_type_to_string(serial)<< " \"\n";
	ofile<<"classname "<< killerdep<<" : \"" <<edge_type_to_string(killerdep)<< " \"\n";
	ofile<<"classname "<< reusedep<<" : \"" <<edge_type_to_string(reusedep)<< " \"\n";
	
	forall_nodes(v, *this){
		it=instruction_type(v);
		bool isvalue=false;
		
		ofile<<"node: { title: \""<<id(v)<< "\"  info1: \""<<it.opcode()<<"\" info2: \" writes " ;
		if (this->is_value(v,t)) {
				isvalue=true;
				ofile<<ISA.get_register_typename(t)<<"(regtype_id="<<t<<") ";
		}	
		
		if(! isvalue){
			ofile <<". lat="<<latency(v)<<", dr="<<delta_r(v);
		}
		ofile <<"\" info3: \""<<get_string_attribute(v)<< "\" color:";
		if(isvalue){
			ofile<<"green}\n";
		}
		else ofile<<"lightgrey}\n";
		
	}
  // unparsing edges in vcg format
	forall_edges(e, *this){
		src = source(e);
		dest = target(e);
		INFO_EDGE ie=inf(e);
		ofile<<"edge: { sourcename: \"";
		ofile<<id(src);
		
		
		ofile<<"\" targetname: \"";
		
		
		ofile<<id(dest);
		
		if (ie.get_register_type_id()<0)
			ofile<<"\" class:"<<etype(e)<<" label:\""<< delta(e)<<"\" color:";
		else
			ofile<<"\" class:"<<etype(e)<<" label:\""<< delta(e)<< "," << this->ISA.get_register_typename(ie.get_register_type_id()) <<"\" color:";
		switch (etype(e)){
			case flowdep_reg: 
				case flowdep_mem: ofile<< "red}\n";
				break;		
			case antidep_reg: 
				case antidep_mem: ofile<<"green}\n";
				break;
			case outputdep_reg: 
				case outputdep_mem: ofile<<"darkmagenta}\n";
				break;	
			case inputdep_reg:
				inputdep_mem :ofile<<"blue}\n";
				break;
				default: serial: ofile<<"black}\n";
		}		
	}// end for edge
	ofile<<"}\n"; 
	
}

int DAG::CriticalPath(){
	return LongestPath();
}
int DAG::CriticalPath(list<node> &nl){
	return LongestPath(nl);
}

int DAG::CriticalPath(list<edge> &el){
	return LongestPath(el);
}

/* Longuest Path of the DAG. return the cost */
int DAG::LongestPath() const
{
	node_array<int> dist(*this);
	node_array<edge> pred(*this), chemin_plus_court(*this);
	edge_array<int> cost(*this);
	LEDA::set<node> S=get_Sources(), //sources
		T=get_Targets();
	LEDA::set<node> down; // to compute reachable nodes from a source
	edge e;
	INFO_EDGE ie;
	bool flag_computed=false; // = true when a longest path has been computed
	int LP = -MAXINT, LP_tmp=-MAXINT; // Longest Paths
	list<edge> le;
	node  v, w;
	
	if (number_of_nodes()==0) return 0; //case of empty DAGs
	/* we give a negative cost to each arc to compute the longuest path
	instead of shortest one */
	
	forall_edges(e, *this){
	ie=inf(e);
	cost[e]=-delta(e);
	}

	
	forall(w, S){
		if (!SHORTEST_PATH_T(*this, w, cost, dist, pred)){
			cerr<<"SIRA error. Longest Path Computation : positive cycle detected in the DAG.\n";
			return EXIT_FAILURE;
		}
		else{
			forall(v, T){
				down=get_down(w);
				if (!down.member(v)) continue; //neglect the case when v is not reachable from w
				le.clear();
				le=COMPUTE_SHORTEST_PATH(*this, w, v, pred); //since cost=-delta, this amounts to longest paths
				LP_tmp=0;
				forall(e, le){ //le contains the longest path from w
					LP_tmp=LP_tmp+delta(e);
				}
				if (LP<LP_tmp){
					LP=LP_tmp;
					flag_computed=true;
				}
			}
		}
			

		
	}
	
	if (flag_computed) return (LP);
	else return EXIT_FAILURE;
}

// this function compute the longest path if no cycle is available.
// the parameter nl contains the node list of the critical path.

int DAG::LongestPath(list<node> &nl) const
{
	node_array<int> dist(*this);
	node_array<edge> pred(*this), chemin_plus_court(*this);
	edge_array<int> cost(*this);
	edge e;
	bool flag_computed=false; // = true when a longest path has been computed
	LEDA::set<node> S=get_Sources(), //sources
		T=get_Targets();
	LEDA::set<node> down; // to compute reachable nodes from a source
	int LP = -MAXINT, LP_tmp=-MAXINT; // Longest Paths
	list<edge> le, le_tmp;

  	node  v, w, u;

  	nl.clear(); // empty nl
	
	forall_edges(e, *this){
	cost[e]=-delta(e);
	}

	forall(w, S){
		
		if (!SHORTEST_PATH_T(*this, w, cost, dist, pred)){
			cerr<<"SIRA error. Longest Path Computation : positive cycle detected in the DAG.\n";
			return EXIT_FAILURE;
		}
		else{
			down=get_down(w);
			forall(v, T){
				if (!down.member(v)) continue; //neglect the case when v is not reachable from w
				SHORTEST_PATH_T(*this, w, cost, dist, pred);
				le_tmp.clear();
				le_tmp=COMPUTE_SHORTEST_PATH(*this, w, v, pred); //since cost=-delta, this amounts to longest paths
				if(!le_tmp.empty()){
					LP_tmp=0;
					forall(e, le_tmp){ //le contains the longest path from w
						LP_tmp=LP_tmp+delta(e);
					}
					if (LP<LP_tmp){
						LP=LP_tmp;
						le=le_tmp;
						flag_computed=true;
					}
				}
			}
		}	
	}
	forall(e, le){
		w=source(e);
		if (nl.search( w )==nil) nl.append(w);
		w=target(e);
		if (nl.search (w)==nil) nl.append(w);
	}
	if(flag_computed) return (LP);
	else return EXIT_FAILURE;
}

int DAG::LongestPath(list<edge> &el) const
{
	node_array<int> dist(*this);
	node_array<edge> pred(*this), chemin_plus_court(*this);
	LEDA::set<node> down; // to compute reachable nodes from a source
	edge_array<int> cost(*this);
	edge e;
	bool flag_computed=false; // = true when a longest path has been computed
	LEDA::set<node> S=get_Sources(), //sources
		T=get_Targets();
	int LP = -MAXINT, LP_tmp=MAXINT; // Longest Paths
	list<edge> le, le_tmp;
	
	node  v, w;
	
	el.clear(); // empty nl
	
	
	forall_edges(e, *this){
		cost[e]=-delta(e);
	}
	
	forall(w, S){
		if (!SHORTEST_PATH_T(*this, w, cost, dist, pred)){
		cerr<<"SIRA error. Longest Path Computation : positive cycle detected in the DAG.\n";
			return EXIT_FAILURE;
		}
		else{
			forall(v, T){
				down=get_down(w);
				if (!down.member(v)) continue; //neglect the case when v is not reachable from w
				le_tmp=COMPUTE_SHORTEST_PATH(*this, w, v, pred); //since cost=-delta, this amounts to longest paths
				LP_tmp=0;
				forall(e, le_tmp){ //le contains the longest path from w
					LP_tmp=LP_tmp+delta(e);
				}
				if (LP<LP_tmp){
					LP=LP_tmp;
					le=le_tmp;
					flag_computed=true;
				}
			}
		}
	}
	el=le; //list of edges belonging to the longest path

	if (flag_computed) return (LP);
	else return EXIT_FAILURE;
}

/*longuest path from top to a node */
int DAG::LongestPathTo(node sink) const{
  return longuest_path_to[sink];
}

/*longuest path from a node to the bottom*/
int DAG::LongestPathFrom(node source) const{
  return longuest_path_from[source];
}
int DAG::ShortestPathFrom(node source) const{
  return shortest_path_from[source];
}
int DAG::ShortestPathTo(node source)const{
  return shortest_path_to[source];
}

set<node> DAG::get_down(node u) const{
	return down[u];
}

set<node> DAG::get_up(node u)const {
	return up[u];
}

set<node> DAG::get_parents(node u) const{
	list<edge> ie=in_edges(u); //in-edges
	set<node> parents;
	edge e;
	forall(e, ie){
		// the source of each in-edge is a parent
		parents.insert(source(e));
	}
	return parents;
}
set<node> DAG::get_children(node u) const{
	list<edge> oe=out_edges(u); //out-edges
	set<node> children;
	edge e;
	forall(e, oe){
		// the source of each in-edge is a parent
		children.insert(target(e));
	}
	return children;
	
}
int DAG::lp(node u, node v) const{
  if(u==v) return 0;
  else{
    return lp_array(u,v);
  }
}
bool DAG::is_value(node n) const {
  INSTRUCTIONS_TYPES it=instruction_type(n);
  return(it.is_value(ISA.get_the_unique_register_type_id()));
}

bool DAG::is_value(node n, int t) const {
  INSTRUCTIONS_TYPES it=instruction_type(n);
  return(it.is_value(t));
}

bool DAG::is_value(int i, int t) const {
  node n;
  INSTRUCTIONS_TYPES it;
  try{
    n=node_with_id(i);
  }
  catch(IllegalAccessNode){
    return false;
  }
  it=instruction_type(n);
  return(it.is_value(t));
}





/* copy a DAG given as argument */
void DAG::copy(const DAG &G){
	node u,v;
	int i,j;
	edge e, ee;
	
	INSTRUCTIONS_TYPES it;
	INFO_EDGE ie;
	this->clear();
	this->set_ISA(G.get_ISA());

	// low level node insertion to keep same node ids
	forall_nodes(u, G){
		it=G.instruction_type(u);
		this->new_node(it.opcode_id(), G.id(u));
	}
	
	//  edge insertion 
	forall_edges(e, G){
		u=G.source(e);
		v=G.target(e);
		new_edge(G.id(u), G.id(v), G[e]);
	}
	  
	build_internal_structures();  
}


void DAG::copy(const DAG &G, node_map<node> &mn, edge_map<edge> &me){
	
	node u,v;
	int i,j;
	edge e, ee;
	mn.init(G);
	me.init(G);
	INSTRUCTIONS_TYPES it;
	INFO_EDGE ie;
	this->clear();
	this->set_ISA(G.get_ISA());

	// low level node insertion to keep same node ids
	forall_nodes(u, G){
		it=G.instruction_type(u);
		i=this->new_node(it.opcode_id(), G.id(u));
		mn[u]=this->node_with_id(i);
	}
	
	//  edge insertion 
	forall_edges(e, G){
		u=G.source(e);
		v=G.target(e);
		ee=new_edge(G.id(u), G.id(v), G[e]);
		me[e]=ee;
	}
	  
	build_internal_structures();  
}

int compare_nodes(const node &u, const node &v) {
	DAG *G1, *G2; 
	list<edge> 	outedges, inedges;
	edge e;
	if (u==NULL){
		cerr<<"DDG Graph error. In function compare_nodes(u,v), u is null.\n";
		return -2;
	}
	if (v==NULL){
		cerr<<"DDG Graph error. In function compare_nodes(u,v), v is null.\n";
		return -2;
	}
	G1=(DAG*) graph_of(u);
	G2= (DAG *) graph_of(v);
	if (G1==NULL){
		cerr<<"DDG Graph error. In function compare_nodes(u,v), G1 is null.\n";
		return -2;
	}
	if (G2==NULL){
		cerr<<"DDG Graph error. In function compare_nodes(u,v), G2 is null.\n";
		return -2;
	}
	if (G1!=G2){
		cerr<<"DDG Graph error. In function compare_nodes(u,v), u and v do not belong to the same graph.\n";
		cerr<<G1->inf(u);
		cerr<<G2->inf(v);
		return -2;
	}
	if(u==v) return 0;

	if (lt(u,v)) return -1;
	
	outedges=G1->out_edges(u);
	forall(e, outedges){
		if (G1->target(e)==v) return (-1); // there exists a path from u to v => u<v
		
	}
	inedges=G1->in_edges(u);
	forall(e, inedges){
		if (G1->source(e)==v) return (+1); // there exists a path from v to u => u>v
	}

	
}

void DAG::get_loop_body(const LOOP &G, node_map<node> &mn, edge_map<edge> &me)
{
    node u,v;
	int i,j;
	edge e, ee;
	set<edge> ER; // flow edges of the current DAG
	clear();
	mn.init(G);
	me.init(G);

	clear();
	
	
	ISA=G.get_ISA(); // first get the same ISA
	
	// insert nodes
	forall_nodes(u, G){
		i=new_node((G.instruction_type(u)).opcode_id(),G.id(u));
		mn[u]=node_with_id(i);
	}
	
	// low level edge insertion to keep same edge information
	forall_edges(e, G){
		if (G.lambda(e)==0){  // only null edges belong to the loop body
			u=G.source(e);
			v=G.target(e);
			ee=new_edge(mn[u],mn[v], G[e]);
			me[e]=ee;	
		}
	}
	build_internal_structures();  
}

void DAG::get_loop_body(const LOOP &G)
{
   	node u,v;
   	int i,j;
	edge e;
	node_array<int> nm(G);
	
	clear();
	
	ISA=G.get_ISA(); // first get the same ISA
	
	// insert nodes
	forall_nodes(u, G){
		nm[u]=new_node((G.instruction_type(u)).opcode_id(),G.id(u));
	}
	
	// low level edge insertion to keep same edge information
	forall_edges(e, G){
		if (G.lambda(e)==0){  // only null edges belong to the loop body
			u=G.source(e);
			v=G.target(e);
			new_edge(nm[u],nm[v], G[e]);
		}
	}
	build_internal_structures();  

}




/*write to a g file */


void DAG::write_to_gl(const char * file_name) const{
	node s, t;
	edge e;
	int nbnode, nbedge, i;
	INSTRUCTIONS_TYPES it;
	ofstream gfile; /* file.g and file.l */
	gfile.open(file_name);
	if(!gfile.is_open()){
		cerr<<"DDG IO failure : could not open " << file_name<<endl;
		return;
	}
	if(ISA.number_of_registers_types()>1){
		cerr<<"DDG warning : writing to gl file while the number of register types is greater than 1.\n";
	}
	
	gfile<<"DDG.GRAPH\n";
	nbnode=number_of_nodes();
	gfile<<nbnode<<endl;
	for(i=1;i<=nbnode;i++){
		try{
			s=node_with_id(i);
		}
		catch(IllegalAccessNode){
			continue;
		}
		it=instruction_type(s);
		gfile<<it.opcode_id()<<endl;
	}
	nbedge=number_of_edges();
	gfile  <<nbedge<<endl;
		
	forall_edges(e,*this){
		gfile<<id(source(e))<<'\t'<<id(target(e))<<'\t'<<delta(e)<<'\t'<<0<<'\t'<< edge_type_to_string(etype(e)) <<endl; // lambda(e)=0 for any edge in a DAG
	}
	
	gfile.close(); 
}

void DAG::write_to_gl(const char * file_name, int regtype_id) const{
	node s, t;
	edge e;
	int nbnode, nbedge, i;
	INSTRUCTIONS_TYPES it;
	ofstream gfile; /* file.g and file.l */
	gfile.open(file_name);
	if(!gfile.is_open()){
		cerr<<"DDG IO failure : could not open " << file_name<<endl;
		return;
	}
	if(ISA.number_of_registers_types()>1){
		cerr<<"DDG warning : writing to gl file while the number of register types is greater than 1.\n";
	}
	
	gfile<<"DDG.GRAPH\n";
	nbnode=number_of_nodes();
	gfile<<nbnode<<endl;
	for(i=1;i<=nbnode;i++){
		try{
			s=node_with_id(i);
		}
		catch(IllegalAccessNode){
			continue;
		}
		it=instruction_type(s);
		gfile<<it.opcode_id()<<endl;
	}
	nbedge=number_of_edges();
	gfile  <<nbedge<<endl;
		
	forall_edges(e,*this){
		if (is_flow_reg(e,regtype_id)){
			gfile<<id(source(e))<<'\t'<<id(target(e))<<'\t'<<delta(e)<<'\t'<<0<<'\t'<< edge_type_to_string(etype(e)) <<endl; // lambda(e)=0 for any edge in a DAG
		}
		else {
			gfile<<id(source(e))<<'\t'<<id(target(e))<<'\t'<<delta(e)<<'\t'<<0<<'\t'<< edge_type_to_string(serial) <<endl; // lambda(e)=0 for any edge in a DAG
		}
	}
	
	gfile.close(); 
}

bool DAG::read_from_gl(const char * file_name){
	node s, t;
	edge e;
	int nbnode, nbedge, i, sourc, targe, delta, lambda=0;
	int tmp;
	DDG::edge_type et;
	std::string bidon;
	INFO_EDGE ie;
	INFO_NODE in;
	char cc[MAX_STRING];
	INSTRUCTIONS_TYPES it;
	list_item li;
	ifstream gfile; /* file.g and file.l */
	gfile.open(file_name);
	clear();
	if(!gfile.is_open()){
		cerr<<"DDG IO Failure: could not open " << file_name<<endl;
		BadIODDG bioddg;
		throw bioddg;
	}
	gfile.getline(cc,MAX_STRING,'\n'); // get tje header section
	bidon=std::string(cc);
	
	std::string str_etype;
	i=bidon.find("DDG.GRAPH"); // looks for the gl header section pecifier
	if ((i<0)||i>bidon.size()) {
		cerr <<"DDG error : bad gl header section : "<<file_name<<". Seems not a valid gl  format (version 1.1).\n";
			gfile.close();
			BadIODDG bioddg;
			throw bioddg;
	}
	gfile>>nbnode;
	for(i=0;i<nbnode;i++){
		gfile>>tmp; /* read the opcode*/
		/* search the opcode in the set of instructions types */
		
		if(!ISA.exist_instruction_type_id(tmp)){
			cerr <<"DDG Error.  Opcode non recognized: "<<tmp<<endl;
			BadIODDG bioddg;
			throw bioddg;
		}
		else {
			new_node(tmp);
		}
	}
	gfile  >>nbedge;
	/* an edge is defined by : source target delta*/
	for(i=0;i<nbedge;i++){
		gfile>>sourc>>targe>>delta>>lambda>>str_etype;
		if(lambda!=0) continue; /* we ignore inter-iteration dep in the case of DAGs */
		try{
			s=node_with_id(sourc);
			t=node_with_id(targe);
		}
		catch(IllegalAccessNode){
			continue;
		}
		try{
			et=string_to_edge_type(str_etype);
		}
		catch(IllegalDependance){
			continue;
		}
		ie=INFO_EDGE(delta, lambda, et);
		if (ie.is_flow_reg()){
			if (ISA.number_of_registers_types()!=1){
				cerr<<"DDG Warning : read from gl while there is not a unique register type in the ISA.\n";
				throw NonUniqueRegisterType();
			}
			ie.set_register_type_id(ISA.get_the_unique_register_type_id());
			
		}
		e=new_edge(s, t, ie);
		
	}
	
	gfile.close();
	build_internal_structures();
	return true;
}


bool DAG::read_from_gl(const char * file_name, int regtype_id){
	node s, t;
	edge e;
	int nbnode, nbedge, i, sourc, targe, delta, lambda=0;
	int tmp;
	DDG::edge_type et;
	std::string bidon;
	INFO_EDGE ie;
	INFO_NODE in;
	char cc[MAX_STRING];
	INSTRUCTIONS_TYPES it;
	list_item li;
	ifstream gfile; /* file.g and file.l */
	gfile.open(file_name);
	clear();
	if(!gfile.is_open()){
		cerr<<"DDG IO Failure: could not open " << file_name<<endl;
		BadIODDG bioddg;
		throw bioddg;
	}
	gfile.getline(cc,MAX_STRING,'\n'); // get tje header section
	bidon=std::string(cc);
	
	std::string str_etype;
	i=bidon.find("DDG.GRAPH"); // looks for the gl header section pecifier
	if ((i<0)||i>bidon.size()) {
		cerr <<"DDG error : bad gl header section : "<<file_name<<". Seems not a valid gl  format (version 1.1).\n";
			gfile.close();
			BadIODDG bioddg;
			throw bioddg;
	}
	gfile>>nbnode;
	for(i=0;i<nbnode;i++){
		gfile>>tmp; /* read the opcode*/
		/* search the opcode in the set of instructions types */
		
		if(!ISA.exist_instruction_type_id(tmp)){
			cerr <<"DDG Error.  Opcode non recognized: "<<tmp<<endl;
			BadIODDG bioddg;
			throw bioddg;
		}
		else {
			new_node(tmp);
		}
	}
	gfile  >>nbedge;
	/* an edge is defined by : source target delta*/
	for(i=0;i<nbedge;i++){
		gfile>>sourc>>targe>>delta>>lambda>>str_etype;
		if(lambda!=0) continue; /* we ignore inter-iteration dep in the case of DAGs */
		try{
			s=node_with_id(sourc);
			t=node_with_id(targe);
		}
		catch(IllegalAccessNode){
			continue;
		}
		try{
			et=string_to_edge_type(str_etype);
		}
		catch(IllegalDependance){
			continue;
		}
		ie=INFO_EDGE(delta, lambda, et);
		if (ie.is_flow_reg()){
			ie.set_register_type_id(regtype_id);
			
		}
		e=new_edge(s, t, ie);		
	}
	
	gfile.close();
	build_internal_structures();
	return true;
}



set<node>  DAG::get_Sources() const{
  return Sources;
}
set<node>  DAG::get_Targets() const{
  return Targets;
}



set<int >   DAG::get_Sources_id()const {
  node n;
  INFO_NODE in;
  set<int> Sources_id;
  set<node> Sources=get_Sources();
  forall(n, Sources){Sources_id.insert((in=inf(n)). id());}
  return Sources_id;
}

set<int >   DAG::get_Targets_id() const {
  node n;
  INFO_NODE in;
  set<int> Targets_id;
  set<node> Targets=get_Targets();
  forall(n, Targets){Targets_id.insert((in=inf(n)). id());}
  return Targets_id;
}
set<int>  DAG::all_nodes_id(){
  node n;
  INFO_NODE in;
  set<int> all_nodes_id;
  forall_nodes(n, *this){all_nodes_id.insert((in=inf(n)). id());}
  return all_nodes_id;

}




int MAXIMAL_ANTI_CHAIN(const graph & G, set<node>& MA){

  node z,y,u,v, zi, zk, xk, xj, zj, yj, yk, yi;
  edge e;
  d_array<int, list<node> > C;
  list<edge> card_match;
  GRAPH<node,edge> G_c; /* Transitive closure of G */
  graph B; /* bipartie graph */
  node_array<node> Y(G, NULL), Z(G, NULL);
  set<node> set_Z, set_Y, unsaturated,  S, T, Tp;
  list<node> Ci;
  list<edge> le;
  list_item item;
  set<node> down;
  node_array<set<node> > G_down(G); /* we associate to each node all its descendants */
  node_array<int> chain(G);
  node_array<node> Z_inverse, couple;
  int max_AM=0, i; /* maximal antichain size */

  /* if this graph is not acyclic ==> return -1 */
  if (!Is_Acyclic(G, le)) return -1;
	if (G.empty()) {
		return 0;
	}	
	
  /* we first begin by building transitive closure to compute descendants for each node */

  G_c=TRANSITIVE_CLOSURE(G);
  forall_edges(e, G_c){
    u=G_c.source(e);
    v=G_c.target(e);
    if(u!=v) G_down[G_c[u]].insert(G_c[v]);
  }
  /* we now construct the bipartie graph B of G
     B(G)=(Y,Z, F) such that
     - Y and Z are two copies of V
     - (y_i,z_i) in F iff (u_i, u_j) in E that is y_i
     (resp z_i) represents x_i in Y (resp z_i in Z) */
  forall_nodes(u, G){
    set_Y.insert(Y[u]=B.new_node());
    set_Z.insert(Z[u]=B.new_node());
  }
  Z_inverse.init(B,NULL);
  /* Z_inverse gives u assigned to z */
  forall_nodes(u,G){
    Z_inverse[Z[u]]=u;
  }
  /* connecte y_i to each zj iff there is a path from xi to xj*/
  forall_nodes(u, G){
    down=G_down[u];
    forall(v, down){
      //      if(u!=v)
	B.new_edge(Y[u], Z[v]);
    }
  }
  couple.init(B, NULL);
  /* we search for the maximal cardinality matching of B
     because it corresponds to a minimum chain partitions */
  card_match=MAX_CARD_BIPARTITE_MATCHING(B, DDG::set_to_list<node> (set_Y), DDG::set_to_list<node>(set_Z));
  forall(e, card_match){
    yi=B.source(e);
    zi=B.target(e);
    couple[yi]=zi;
    couple[zi]=yi;
  }
  /* antichain max = number of minimal chains */

  /* now we construct the minimal chain partition */
  forall(zj, set_Z){
    if (couple[zj]==NULL){ /* we have a source of a new chain */
      unsaturated.insert(Z_inverse[zj]);
      max_AM++;
      Ci.clear();
      Ci.append(xj=Z_inverse[zj]);
      chain[xj]=max_AM;
      yj=Y[xj];
      while(couple[yj]!=NULL){
	zk=couple[yj];
	xk=Z_inverse [zk];
	Ci.append(xk);
	chain[xk]=max_AM;
	yj=Y[xk];
      }
      C[max_AM]=Ci;
    }
  }
 
  /* now we have the minimal partition in C
     to compte a maximal antichain, we select an element from
     each chain such that it is parallel with the others */
  S = unsaturated;
  do {

    T.clear();
    /*T is the set of nodes to be changed */
    forall(u, S){
      down=G_down[u];
      forall(v, S){
	if((u!=v) && (down.member(v))){ /* u should be removed from the max antichain */
	  T.insert(u);
	  break;
	}
      }
    }

    /* Tp are nodes to be added */
    Tp.clear();
    forall(u, T){
      Ci=C[chain[u]];
      if((item=Ci.succ(Ci.search(u)))!=NULL)
	Tp.insert(Ci[item]);
    }

    
    /* try now with this */
    S=S-T+Tp;
  } while(!T.empty());
  
  /* the maximal antichain is S */
  MA =S;
  return max_AM;
}

int MINIMAL_CHAIN(const graph &G, node_array<int>& chain){
	node z,y,u,v, zi, zk, xk, xj, zj, yj, yk, yi;
	edge e;
  	list<edge> card_match;
  	GRAPH<node,edge> G_c; /* Transitive closure of G */
	graph B; /* bipartie graph */
	node_array<node> Y(G, NULL), Z(G, NULL);
	set<node> set_Z, set_Y, unsaturated,  S, T, Tp;
	list<node> Ci;
	list<edge> le;
	list_item item;
	set<node> down;
	node_array<set<node> > G_down(G); /* we associate to each node all its descendants */
	node_array<node> Z_inverse, couple;
	int i=0; /* number of chains */

	chain.init(G);
	/* if this graph is not acyclic ==> return -1 */
	if (!Is_Acyclic(G, le)) {
		cerr<<"DDG error. Dilworth decomposition cannot be done on cyclic graph"<<endl;
		return -1;
	}
	if (G.empty()) {
		return 0;
	}

	/* we first begin by building transitive closure to compute descendants for each node */

	G_c=TRANSITIVE_CLOSURE(G);
	forall_edges(e, G_c){
		u=G_c.source(e);
		v=G_c.target(e);
		if(u!=v) G_down[G_c[u]].insert(G_c[v]);
	}
	/* we now construct the bipartie graph B of G
	B(G)=(Y,Z, F) such that
	- Y and Z are two copies of V
	- (y_i,z_i) in F iff (u_i, u_j) in E that is y_i
	(resp z_i) represents x_i in Y (resp z_i in Z) */
	forall_nodes(u, G){
	set_Y.insert(Y[u]=B.new_node());
	set_Z.insert(Z[u]=B.new_node());
	}
	Z_inverse.init(B,NULL);
	/* Z_inverse gives u assigned to z */
	forall_nodes(u,G){
	Z_inverse[Z[u]]=u;
	}
	/* connecte y_i to each zj iff there is a path from xi to xj*/
	forall_nodes(u, G){
	down=G_down[u];
		forall(v, down){
		//      if(u!=v)
		B.new_edge(Y[u], Z[v]);
		}
	}
	couple.init(B, NULL);
	/* we search for the maximal cardinality matching of B
	because it corresponds to a minimum chain partitions */
	card_match=MAX_CARD_BIPARTITE_MATCHING(B, set_to_list<node> (set_Y), set_to_list<node>(set_Z));
	forall(e, card_match){
		yi=B.source(e);
		zi=B.target(e);
		couple[yi]=zi;
		couple[zi]=yi;
	}
	/* antichain max = number of minimal chains */

	/* now we construct the minimal chain partition */
	
	forall(zj, set_Z){
		if (couple[zj]==NULL){ /* we have a source of a new chain */
			unsaturated.insert(Z_inverse[zj]);
			Ci.clear();
			Ci.append(xj=Z_inverse[zj]);
			chain[xj]=i;
			yj=Y[xj];
			while(couple[yj]!=NULL){
					zk=couple[yj];
					xk=Z_inverse [zk];
					Ci.append(xk);
					chain[xk]=i;
					yj=Y[xk];
			}
			i++;
		}
	}
	
	/* now we have the minimal partition in chain*/
	return (i);
}


int MINIMAL_CHAIN(const DAG &G, array< list<node> > &C){
	node_array<int> chain;
	int p, i;
	node u;
	edge e;
	p=MINIMAL_CHAIN(G, chain); // dilworth decomposition of the DAG
	if(p<=0) return p;
	
	C.resize(p);
	
	forall_nodes(u,G){
		C[chain[u]].append(u);
	}

	for(i=0;i<p;i++){
		C[i].sort(&compare_nodes); // sort the nodes inside the chains
	}
	return p;
}


}


int DDG::DAG::max_delta(node u, node v) const
{
	edge e;
	int i=-MAXINT,j;
	set<edge> SE=edges_from_to(u,v);
	forall(e,SE){
		if((j=delta(e))>i){
			i=j;
		}
	}
	return i;
}

int DDG::DAG::min_delta(node u, node v) const
{
	edge e;
	int i=MAXINT,j;
	set<edge> SE=edges_from_to(u,v);
	forall(e,SE){
		if((j=delta(e))<i){
			i=j;
		}
	}
	return i;
}


DDG::INSTRUCTIONS_TYPES DDG::DAG::search_instruction_type(const std::string opcode) const
{
	list_item li;
	INSTRUCTIONS_TYPES it;
	if (ISA.exist_instruction_type_opcode(opcode)){
		it=ISA.get_instruction_type_by_opcode(opcode);
		return it;
	}
	
	// if not found
	it.set_opcode_id(-1);
	return it;
}

DDG::INSTRUCTIONS_TYPES DDG::DAG::search_instruction_type(int opcode_id) const
{
	list_item li;
	INSTRUCTIONS_TYPES it;
	if (ISA.exist_instruction_type_id(opcode_id)){
		it=ISA.get_instruction_type_by_id(opcode_id);
		return it;
	}	
	it.set_opcode_id(-1);
	return it;
}


/// @todo implement me Frederic
void DDG::DAG::write_to_xml(const  char *filename){
  ofstream fp;

  try{
    fp.open(filename);

    fp << "<ddg version='1'>\n";

    //@todo : take care of hidden nodes/edges
    
    // First, list all nodes
    const LEDA::list<node>& nodes = all_nodes();
    node n;
    forall(n, nodes){
      std::string type = instruction_type(n).opcode();
      int my_id = id(n);
      std::string description = get_string_attribute(n); 
      fp << "<operation type='" << type <<"' id='" << my_id << "'";
      if(description.length() > 0){
        // Escape the double quotes
        std::string str2("\\");
        int position = -2;
        while((position = description.find("\"", position+2)) != std::string::npos){
          description.insert(position, str2);
        }
        fp << " description=\"" << description << "\"";
      }        
      fp << "/>\n";
    }

    // Then list all edges
    const LEDA::list<edge>& edges = all_edges();
    edge e;
    forall(e, edges){
      int src = id(source(e));
      int dst = id(target(e));
      int latency     = inf(e).delta();
      int distance    = inf(e).lambda();
      DDG::edge_type et    = inf(e).etype();
      fp << "<edge src='" << src << "' dst='" << dst << "' latency='" << latency << "' distance='" << distance << "' typedep='" << edge_type_to_string(et) << "'";
      switch(et){
          case flowdep_reg :
          case antidep_reg :
          case outputdep_reg :
          case inputdep_reg :
          case killerdep :
          case reusedep :
            std::string regtype = ISA.get_register_typename(get_register_type_id(e));
            fp << " regtype='" << regtype << "'";
      }
      fp << "/>\n";
    }
    fp << "</ddg>\n";
    fp.close();
  }catch( ... ){
    cerr << "DDG error encountered while writing DAG to xml\n";
  }
}


/// @todo implement me Frederic
bool DDG::DAG::read_from_xml(const char *xml_file)
{
  int result = -1;
  
  try{
    XMLPlatformUtils::Initialize();
  } catch ( ... ){
    cerr << "DDG - Error while initializing XML" << xml_file<<endl;
    return false;
  }
  this -> clear();
  xercesc::XercesDOMParser *config_parser = new XercesDOMParser;
  // Tags and attributes we are interested in parsing.
  XMLCh* TAG_DDG        = XMLString::transcode("ddg");
  XMLCh* TAG_Operation  = XMLString::transcode("operation");
  XMLCh* TAG_Edge       = XMLString::transcode("edge");
  
  XMLCh* ATTR_Version     = XMLString::transcode("version");
  XMLCh* ATTR_Distance    = XMLString::transcode("distance");
  XMLCh* ATTR_Latency     = XMLString::transcode("latency");
  XMLCh* ATTR_Regtype     = XMLString::transcode("regtype");
  XMLCh* ATTR_Src         = XMLString::transcode("src");
  XMLCh* ATTR_Dst         = XMLString::transcode("dst");
  XMLCh* ATTR_Type        = XMLString::transcode("type");
  XMLCh* ATTR_Typedep     = XMLString::transcode("typedep");
  XMLCh* ATTR_Description = XMLString::transcode("description");
  XMLCh* ATTR_Id          = XMLString::transcode("id");
  
  try{
    // Check if file exists and can be read
    struct stat filebuf;
    int status;
    status = stat(xml_file, &filebuf);
    
    if (status != 0){
      cerr << "DDG error: Cannot stat XML file" << xml_file<<endl;
      InvalidISADesc inv;
      throw (inv);
    }
    
    // @todo : use DTD
    config_parser->setValidationScheme( XercesDOMParser::Val_Auto);
    
    config_parser->parse(xml_file);
    
    DOMDocument* xml_doc = config_parser->getDocument();
    
    // Get the top-level element : "arch"
    DOMElement* root = xml_doc->getDocumentElement();
    if( !root ){
      cerr << "DDG error: Error with input XML file" << xml_file<<endl;
      InvalidISADesc inv;
      throw (inv);
    }
    
    int version  = XMLString::parseInt(root->getAttribute(ATTR_Version));
    
    // @todo Only support version 1
    if(version != 1){
      cerr << "Wrong XML version " << xml_file<<endl;
      InvalidISADesc inv;
      throw (inv);
    }

    // First, operations 
    DOMNodeList* operations = root->getElementsByTagName(TAG_Operation);
    XMLSize_t node_count = operations->getLength();
    
    for( XMLSize_t x = 0; x < node_count; ++x ) {
      DOMNode* current_node = operations->item(x);
      if( current_node->getNodeType() && current_node->getNodeType() == DOMNode::ELEMENT_NODE ){
        DOMElement* current_element = dynamic_cast< xercesc::DOMElement* >( current_node );
        
        // Create the new operation 
        char *type = XMLString::transcode(current_element->getAttribute(ATTR_Type));
        int id     = XMLString::parseInt(current_element->getAttribute(ATTR_Id));
        char *description = XMLString::transcode(current_element->getAttribute(ATTR_Description));
        std::string stype = std::string(type);
        if(!ISA.exist_instruction_type_opcode(stype)){
          cerr <<"DDG Error.  Opcode non recognized: "<< stype<< " in "<< xml_file<<endl;
          BadIODDG bioddg;
          throw bioddg;
        }
        new_node(ISA.get_instruction_type_by_opcode(stype).opcode_id(), id, std::string(description));
        XMLString::release(&description);
        XMLString::release(&type);
      }
    }
    
    // 
    DOMNodeList* edges = root->getElementsByTagName(TAG_Edge);
    node_count = edges->getLength();
    
    for( XMLSize_t x = 0; x < node_count; ++x ) {
      DOMNode* current_node = edges->item(x);
      if( current_node->getNodeType() && current_node->getNodeType() == DOMNode::ELEMENT_NODE ){
        DOMElement* current_element = dynamic_cast< xercesc::DOMElement* >( current_node );
        
        // Create the new edge 
        int src      = XMLString::parseInt(current_element->getAttribute(ATTR_Src));
        int dst      = XMLString::parseInt(current_element->getAttribute(ATTR_Dst));
        char *typedep  = XMLString::transcode(current_element->getAttribute(ATTR_Typedep));
        int latency  = XMLString::parseInt(current_element->getAttribute(ATTR_Latency));
        int distance = XMLString::parseInt(current_element->getAttribute(ATTR_Distance));
       
        if(distance != 0){
          // Sid -> Fred : ici ce n'est pas une erreur, on n�glige l'arc c'est tout.
          //cerr << "Warning - DAG with distance != 0\n";
          // throw IllegalDependance();
          continue;
        }
        
        node s = node_with_id(src);
        node t = node_with_id(dst);
        DDG::edge_type et = string_to_edge_type(std::string(typedep));
        INFO_EDGE ie = INFO_EDGE(latency, distance, et);
        switch(et){
          case flowdep_reg :
          case antidep_reg :
          case outputdep_reg :
          case inputdep_reg :
          case killerdep :
          case reusedep :
            char *reg_type = XMLString::transcode(current_element->getAttribute(ATTR_Regtype));
            int id = ISA.get_register_type_id(std::string(reg_type));
            if(id == -1){
              cerr << "DDG error. Bad register type" << std::string(reg_type) << "in"<< xml_file<< endl;
              throw InvalidISADesc();
            }
            ie.set_register_type_id(id);
            XMLString::release(&reg_type);
        }
        XMLString::release(&typedep);
        new_edge(s, t, ie);
      }
    }
    result = 1;
    build_internal_structures();
    
  }catch( ... ){
    cerr<<"DDG - Error while parsing XML file "<<xml_file<<endl;
    result = -1;
  }
  
  // Clean up
  XMLString::release(&TAG_DDG       );
  XMLString::release(&TAG_Operation );
  XMLString::release(&TAG_Edge      );
  XMLString::release(&ATTR_Version  );
  XMLString::release(&ATTR_Distance );
  XMLString::release(&ATTR_Latency  );
  XMLString::release(&ATTR_Regtype  );
  XMLString::release(&ATTR_Src      );
  XMLString::release(&ATTR_Dst      );
  XMLString::release(&ATTR_Type     );
  XMLString::release(&ATTR_Typedep  );
  XMLString::release(&ATTR_Description);
  XMLString::release(&ATTR_Id);
  XMLPlatformUtils::Terminate();
  
  return true;
}

void DDG::DAG::set_string_attribute(node n, std::string inst_text){
	INFO_NODE in=inf(n);
	in.set_string_attribute(inst_text);
	LEDA::GRAPH<INFO_NODE, INFO_EDGE>::assign(n, in);
}

std::string DDG::DAG::get_string_attribute(node n) const
{
	INFO_NODE in=inf(n);
	return in.get_string_attribute();
}

void DDG::DAG::set_delta(edge e, int l)
{
	INFO_EDGE ie=inf(e);
	ie.set_delta(l);
	LEDA::GRAPH<INFO_NODE, INFO_EDGE>::assign(e, ie);
}

void  DDG::DAG::set_instruction_type(node n, int it_id)
{
	INSTRUCTIONS_TYPES it=ISA.get_instruction_type_by_id(it_id);

	INFO_NODE in=inf(n);
	in.set_instruction_type(it);
	LEDA::GRAPH<INFO_NODE, INFO_EDGE>::assign(n, in);
}

void  DDG::DAG::set_etype(edge e, DDG::edge_type et)
{
	INFO_EDGE ie=inf(e);
	ie.set_etype(et);
	LEDA::GRAPH<INFO_NODE, INFO_EDGE>::assign(e, ie);
}

void DDG::DAG::hide_edges(const LEDA::list<edge>& el)
{ edge e;
	forall(e, el)	hide_edge(e);
}

void DDG::DAG::restore_edges(const LEDA::list<edge>& el)
{ edge e;
	forall(e, el) restore_edge(e);
}

void DDG::DAG::hide_nodes(const LEDA::list<node>& nl)
{ node u;
	forall(u, nl) hide_node(u);
}

DDG::edge_type DDG::DAG::etype(edge e) const
{
	INFO_EDGE ie=inf(e);
	return ie.etype();
}

int DDG::DAG::delta(edge e) const
{
	INFO_EDGE ie=inf(e);
	return ie.delta();
}

int DDG::DAG::latency (node n) const
{
	INSTRUCTIONS_TYPES it=instruction_type(n);
	return it.latency();
}

int DDG::DAG::delta_r (node n)const {
	INSTRUCTIONS_TYPES it=instruction_type(n);
	return it.delta_r();
}

int DDG::DAG::delta_r (int idn) const{
	INSTRUCTIONS_TYPES it=instruction_type(idn);
	return it.delta_r();
}

int DDG::DAG::delta_w (node n) const{
	INSTRUCTIONS_TYPES it=instruction_type(n);
	if (ISA.number_of_registers_types()!=1){
		cerr<<"DDG error: call of delta_w while there is no unique register type";
		throw NonUniqueRegisterType();
	}
	return it.delta_w(ISA.get_the_unique_register_type_id());
}

int DDG::DAG::delta_w (int idn) const{
	INSTRUCTIONS_TYPES it=instruction_type(idn);
	
	if (ISA.number_of_registers_types()!=1){
		cerr<<"DDG error: call of delta_w while there is no unique register type";
		throw NonUniqueRegisterType();
	}
	return it.delta_w(ISA.get_the_unique_register_type_id());
}

int DDG::DAG::delta_w (node n, int regtype_id) const{
	INSTRUCTIONS_TYPES it=instruction_type(n);
	if (! ISA.exist_register_type_id(regtype_id)){
		cerr<<"DDG error: the register type id "<< regtype_id<<" does not exist.\n";
		throw BadRegType();
	}
	return it.delta_w(regtype_id);
}

int DDG::DAG::delta_w (int idn, int regtype_id) const{
	INSTRUCTIONS_TYPES it=instruction_type(idn);
	
	if (! ISA.exist_register_type_id(regtype_id)){
		cerr<<"DDG error: the register type id "<< regtype_id<<" does not exist.\n";
		throw BadRegType();
	}
	return it.delta_w(regtype_id);
}


DDG::INSTRUCTIONS_TYPES DDG::DAG::instruction_type(node n)  const
{
	INFO_NODE in;
	in=this->inf(n);
	return in.instruction_type();
}

DDG::INSTRUCTIONS_TYPES DDG::DAG::instruction_type(int idn)  const
{
	node n=node_with_id(idn);
	INFO_NODE in;
	in=this->inf(n);
	return in.instruction_type();
}






bool DDG::lt(node u, node v){
	DAG *G1=(DAG *) graph_of(u), *G2= (DAG *) graph_of(v);
	set<node> down;
	if ((G1==NULL)||(G2==NULL)){
		cerr<<"DDG error. In function lt(u,v), one of the graphs is null.\n";
		return false;
	}
	if (G1!=G2){
		cerr<<"DDG error. In function lt(u,v), u and v do not belong to the same graph.\n";
		return false;
	}
	down=G1->get_down(u);
	if((u!=v)&&(down.member(v))) return true;
	else return false;
}


bool DDG::le(node u, node v){
	DAG *G1=(DAG *) graph_of(u), *G2= (DAG *) graph_of(v);
	set<node> down;
	if ((G1==NULL)||(G2==NULL)){
		cerr<<"DDG error. In function le(u,v), one of the graphs is null.\n";
		return false;
	}
	if (G1!=G2){
		cerr<<"DDG error. In function le(u,v), u and v do not belong to the same graph.\n";
		return false;
	}
	down=G1->get_down(u);
	if(down.member(v)) return true;
	else return false;
}


bool DDG::gt(node u, node v){
	DAG *G1=(DAG *) graph_of(u), *G2= (DAG *) graph_of(v);
	set<node> up;
	if ((G1==NULL)||(G2==NULL)){
		cerr<<"DDG error. In function gt(u,v), one of the graphs is null.\n";
		return false;
	}
	if (G1!=G2){
		cerr<<"DDG error. In function gt(u,v), u and v do not belong to the same graph.\n";
		return false;
	}
	up=G1->get_up(u);
	if((u!=v)&&up.member(v)) return true;
	else return false;
}


bool DDG::ge(node u, node v){
	DAG *G1=(DAG *) graph_of(u), *G2= (DAG *) graph_of(v);
	set<node> up;
	if ((G1==NULL)||(G2==NULL)){
		cerr<<"DDG error. In function gt(u,v), one of the graphs is null.\n";
		return false;
	}
	if (G1!=G2){
		cerr<<"DDG error. In function gt(u,v), u and v do not belong to the same graph.\n";
		return false;
	}
	up=G1->get_up(u);
	if(up.member(v)) return true;
	else return false;
}

bool DDG::parallel(node u, node v){
	if (u==v) return false;
	return (!lt(u,v)&& !lt(v,u));
}

bool DDG::comparable(node u, node v){
	if (u==v) return true;
	return (lt(u,v) || lt(v,u));
}



/*!
    \fn DDG::DAG::get_register_type()
 */
int DDG::DAG::get_register_type_id(edge e) const
{
    INFO_EDGE ie=inf(e);
    return ie.get_register_type_id();
}


/*!
    \fn DDG::DAG::check()
 */
bool DDG::DAG::check()
{	INSTRUCTIONS_TYPES it;
	edge e;
    /// @todo implement me
	//--check edges latencies vs operations latencies
	forall_edges(e, *this)
	{
		if (latency(source(e)) > lp(source(e), target(e))){
				 it=instruction_type(source(e));
			cerr<<"DDG warning : the latency of the operation "<<id(source(e))<<" of type "<<it.opcode()<<" is larger than the longest path between the nodes "<<id(source(e))<<" and "<<id(target(e))<<endl;
			}
	}
	
    return true;
}

bool DDG::DAG::check(int t)
{
    /// @todo implement me
	return true;
}




/*!
    \fn DDG::DAG::set_node_flag ( const NodeFlag& theValue )
 */
void DDG::DAG::set_node_flag ( node n, NodeFlag theValue )
{
    INFO_NODE in=this->inf(n);
    in.set_node_flag(theValue);
    LEDA::GRAPH<INFO_NODE, INFO_EDGE>::assign(n, in);
}


/*!
    \fn DDG::DAG::node_flag ( node n) const;
 */
DDG::NodeFlag DDG::DAG::node_flag ( node n) const
{
   return (this->inf(n)).get_node_flag();
}


/*!
    \fn DDG::DAG::T() const 
    This method returns the list of register types used in the DDG. The resulting T list exclused any register type t such that \f$ V_{R,t}=\phi \f$.
 */
LEDA::list<DDG::REGISTER_TYPES> DDG::DAG::T() const 
{
	LEDA::list<REGISTER_TYPES> TT=this->ISA.T();
	LEDA::set<node> VR;
	REGISTER_TYPES rt;
	forall(rt, TT){
		VR=this->get_V_R(rt.get_register_type_id());
		if (VR.empty()) TT.remove(rt);
	}
	return TT;
}
