/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/
 
/*! \file RS_example.cpp
\brief Register Saturation Computation
This is an example to show how to compute the register saturation of a DAG for each register type.
*/
#include "DDG.h"
#include <iostream>
#include <cstdlib>


using namespace std;
using namespace DDG; // the namespace of the DDG library.



int main(int argc, char *argv[])
{

	// some strings for file names
	
	char        str[MAX_STRING];
	
	// some objects for graphs
	DDG::DAG dag;  // declare a DAG of a basic block (directed acyclic graph)
	DDG::ARCHITECTURE isa_arch;
	char        *ddg_filename=NULL, *arch_filename=NULL, *regtypename=NULL;
	LEDA::node n,u;
	LEDA::edge e;
	int c,i,t;
	LEDA::list<node> nl;
	INSTRUCTIONS_TYPES it;
	LEDA::set<node> MA; // for a maximal antichain computation
	LEDA::set<node> VR; // Set of values
	LEDA::node_array<node> killer;
	REGISTER_TYPES rt;
	LEDA::list<REGISTER_TYPES> T;

	while ((c = getopt (argc, argv, "i:a:t:")) != -1){
		switch(c){
			case 'a': arch_filename=optarg;
				break;
			case 'i': ddg_filename =optarg;
				break;
			case 't': regtypename=optarg;
				break;
			case '?': 
			 	cout<<"usage:"<< argv[0] <<" -i ddg_filename.xml [-a isa_desc.xml] [-t register_type_name] "<<endl;
			 	return EXIT_FAILURE;
		}
		
	}
	if (arch_filename!=NULL)  { 
		isa_arch.read_from_xml(arch_filename);
		if(isa_arch.check()){
			dag=DAG(isa_arch);	 //build an empty DDG loop with a user defined architcural (ISA) description
 		} else return EXIT_FAILURE;
		
	} 
	 //---- The user can read a DDG from an xml format defined by the DDG library
	if (ddg_filename != NULL){
		list<edge> le;
		edge e;
		i=dag.read_from_xml(ddg_filename); // reads the DDG loop from an xml file
		if(i==-1) return EXIT_FAILURE;
		if(! Is_Acyclic(dag,le)==true) {
			cerr<<"Input Error: Non Acyclic graph. Illegal arcs:\n";
			forall(e, le){
				cerr<<dag.id(dag.source(e))<<" -> "<< dag.id(dag.target(e))<<endl;
			}
			return EXIT_FAILURE;
		}

	}
	else{
	    cout<<"usage:"<< argv[0] <<" -i ddg_filename.xml [-a isa_desc.xml][-t register_type_name] "<<endl;
	    return EXIT_FAILURE;
	}
	
	if (regtypename!=NULL){
		
		if(!isa_arch.exist_register_typename(regtypename)){
			cerr<<"Register type of name "<<regtypename<<" does not exist in the ISA.\n";
			return EXIT_FAILURE;
		}
		t=isa_arch.get_register_type_id( regtypename );
		cout<<"Register Saturation of type "<< regtypename<<" of the DAG (loop body) of "<<ddg_filename<<" is : "
		<<REGISTER_SATURATION (dag, t, nl, killer)<<endl;
		cout<<"Saturating values of type "<< regtypename <<endl;
		forall(n, nl){
			it= dag.instruction_type(n);
			cout<< dag.id(n)<< " with opcode " << it.opcode()<<endl;
		}
		cout<<"Saturating killing function of type "<< regtypename <<endl;
		VR=dag.get_V_R(t);
		forall(n, VR){
		if(killer[n]!=NULL) 	cout<<"k("<<dag.id(n)<<")="<<dag.id(killer[n])<<endl;
		else cout<<"k("<<dag.id(n)<<")= bottom"<<endl;
		}
		return EXIT_SUCCESS;
	}

	T=dag.T();
	
	
	forall(rt, T) {
		t=rt.get_register_type_id();
		cout<<"Register Saturation of type "<< rt.get_register_typename()<<" of the DAG (loop body) of "<<ddg_filename<<" is : "
		<<REGISTER_SATURATION (dag, t, nl, killer)<<endl;
		cout<<"Saturating values of type "<< rt.get_register_typename() <<endl;
		forall(n, nl){
			it= dag.instruction_type(n);
			cout<< dag.id(n)<< " with opcode " << it.opcode()<<endl;
		}
		cout<<"Saturating killing function of type "<< rt.get_register_typename() <<endl;
		VR=dag.get_V_R(t);
		forall(n, VR){
		if(killer[n]!=NULL) 	cout<<"k("<<dag.id(n)<<")="<<dag.id(killer[n])<<endl;
		else cout<<"k("<<dag.id(n)<<")= bottom"<<endl;
		}
	}
	return EXIT_SUCCESS;
}
