/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/


/*! \file check_lexicographic_positive.cpp
\brief A program example that shows how to check if a loop DDG is lexicographic positive. It uses an optimised retiming method based on the fundamental results of Leiserson and Saxe published in Algorithmica.
Details.
*/

#include "DDG.h" // the header file of the DDG library. To be included by the user.

#include <iostream>
#include <cstdlib>


using namespace std;
using namespace DDG; // the namespace of the DDG library.

char *getStringForArgument(char * s, int argc, char *argv[])
 {
	for (int i = 1 ; i<argc; i++)
	{
		if (strcmp(argv[i],s) == 0)
		{ 
			if (i+1<argc){
			return argv[i+1];}
			else
			fprintf(stderr,"Argument value missing for %s\n",s);
			return(NULL);
		}
	}
	return(NULL);
}


int main(int argc, char *argv[])
{
	//int are always usefull
	int         i,j ;
	// some strings for file names
	ARCHITECTURE isa_arch;
	char        *ifi;
	char        cc[MAX_STRING], str[MAX_STRING];
	
	
	// some objects for graphs
	//declare a DDG of a loop (cyclic data dependence graph)
	DDG::LOOP loop;

	edge e;
	
	
	
	ifi = getStringForArgument("-i",argc,argv);
	if (ifi==NULL)  { // no line commande arguments
		cout<<"usage"<< argv[0]<<"-i ddg_filename.xml [-a isa_desc.xml] "<<endl;
		return EXIT_SUCCESS;
	}

	if (ifi==NULL){
		cout << "specify a .xml file"<<endl;
		exit(1);
	}
	strcpy(cc,ifi);
	//------- the user can define its own ISA ----------------
	ifi = getStringForArgument("-a",argc,argv);
	//check wether the user specifies an ISA description or not
	if (ifi!=NULL)  {
		//build an empty DDG loop with a user defined ISA description
		isa_arch.read_from_xml(ifi);
		if(isa_arch.check())	loop=LOOP(isa_arch);
		else exit(1);
	}
	 
	 //---- The user can read a DDG from a gl format defined by the DDG library

	loop.read_from_xml(cc); // reads the DDG loop from a gl file
	

	if(is_lexicographic_positive(loop, e)){
		cout<<cc<< ": This DDG is lexicographic positive.\n";
	}
	else {
		cout<<cc<< ": This DDG is not lexicographic positive.\n";
		cout<<" The edge "<< loop.source_id(e) <<" -> " <<loop.target_id(e)<<" with distance equal to "<<loop.lambda(e)<< " belongs to a non positive circuit.\n";
	}

	return EXIT_SUCCESS;
}

