/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/

#ifndef __ddg_exceptions_H
#define __ddg_exceptions_H

#include <iostream>
#include <fstream>

/*! \file ddg_exceptions.h
\brief All exception objects thrown by DDG library.
DDG objects and methods may throw exceptions. In the documentation of each method, we precise the thrown exception if any. 
*/
using namespace std;
namespace DDG{
/* Exceptions */

  class IllegalAccessNode{
	  public:
		IllegalAccessNode(int i){
 		 cerr<<"DDG runtime exception. Illegal node access: the node with id " << i <<" doesn't exists"<<endl;
		}
		  IllegalAccessNode();
  };

  class InvalidISADesc{
   public:
   InvalidISADesc(){
   	cerr<<"DDG exception. Invalid ISA description.\n";
   }
  };
  class BadIODDG{
  	public:
  	BadIODDG(){
  	cerr<<"DDG exception. Bad Input DDG.\n";
  	}
  };
  
  class BadRegType{
  	public:
  	BadRegType(){
  		cerr<<"DDG exception. Bad register type.\n";
  	}
  };
  class BadInstructionType{
  	public:
  	BadInstructionType(){
  	cerr<<"DDG exception. Bad instruction type.\n";
  	}
	BadInstructionType(int id){
		cerr<<"DDG exception. Bad instruction type id "<<id<<endl;
	}
	
  };
  
  class NonUniqueRegisterType{
  	public:
  	NonUniqueRegisterType(){
  	cerr<<"DDG exception. Non unique register type.\n";
  	}
  };
  
  class IllegalDependance{
  	public:
  	IllegalDependance(){
  		cerr<<"DDG exception. Illegal Dependance type.\n";
  	}
  };
  
  class NonUniqueNodeID{
  	public:
  	NonUniqueNodeID(int u){
  		
  	cerr<<"DDG  exception (new node insertion) : a node with id " << u <<" already exists"<<endl;
	}
	NonUniqueNodeID(){
	 cerr<<"DDG  exception (new node insertion) : a node already exists.\n";
	}
  	
  };
  
}

#endif
