/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/

/*! \file info_node.cpp
\brief Implementations of the methods of INFO_NODE class.
Each node in a DDG has multiple attributes. The current file provides the detailed implementations of all methods accessing/modifying them.
*/
#include "info_node.h"

namespace DDG{

INFO_NODE::INFO_NODE() : text(), optype()
{ 
	identifier=-1;
	node_flag=NodeFlag_Nothing;
}

INFO_NODE::INFO_NODE(int id, INSTRUCTIONS_TYPES opt) : text()
{
	identifier=id;
	optype=opt;
	node_flag=NodeFlag_Nothing;
}

INFO_NODE::INFO_NODE(const INFO_NODE &in): text()
{
	identifier=in.identifier;
	optype=in.optype;
	text=in.text;
	node_flag=in.node_flag;
}

INFO_NODE::INFO_NODE(int id, INSTRUCTIONS_TYPES opt, std::string inst_text): text()
{
	identifier=id;
	optype=opt;
	text=inst_text;
	node_flag=NodeFlag_Nothing;
}

ostream& operator<<(ostream &os,  INFO_NODE x) /* output stream */
{
	std::string str_attr=x.get_string_attribute();
	if(str_attr.empty()){
		os<<x.id()<<' '<<x.instruction_type()
			<<' '; //textual attribute written between double quotes
	}
	else {
		//replace newline in textual attributes by white space
		replace(str_attr.begin(), str_attr.end(), '\n', ' ');
		os<<x.id()<<' '<<x.instruction_type()
		<<' '<<str_attr; 
	}
	return os;
}

istream& operator>>(istream &is, INFO_NODE & x) /* input stream */
{
	int i;
	std::string text_attr;
	INSTRUCTIONS_TYPES it;
	is>>i>>it;
	getline(is,text_attr);//read textual attribute
	//replace newline in textual attributes by white space
	replace(text_attr.begin(), text_attr.end(), '\n', ' ');

	x.set_id(i);
	x.set_instruction_type(it);
	x.set_string_attribute(text_attr);
	return is;
}

/*!
    \fn DDG::INFO_NODE::check()
    \returns true if the object seems ok.
Current implementation always return true.
 */
bool INFO_NODE::check()
{
    /// @todo implement me
	return true;
}
}
