/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/

/*! \file loop.cpp
\brief Contains the implementations of LOOP methods.
*/
#include "loop.h"
#include "LEDA/core/map2.h" //map2
#include "LEDA/graph/shortest_path.h" /*belman ford */

#include <xercesc/dom/DOM.hpp>
#include <xercesc/dom/DOMDocument.hpp>
#include <xercesc/dom/DOMDocumentType.hpp>
#include <xercesc/dom/DOMElement.hpp>
#include <xercesc/dom/DOMImplementation.hpp>
#include <xercesc/dom/DOMImplementationLS.hpp>
#include <xercesc/dom/DOMNodeIterator.hpp>
#include <xercesc/dom/DOMNodeList.hpp>
#include <xercesc/dom/DOMText.hpp>

#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/util/XMLUni.hpp>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

XERCES_CPP_NAMESPACE_USE
        
using namespace LEDA;
using namespace std;

namespace DDG {


	
LOOP::LOOP()
	:DAG()
{	
}
LOOP::LOOP(ARCHITECTURE isa )
	:DAG(isa)
{			
}


LOOP::~LOOP()
{
}

void LOOP::copy(const LOOP &G) {
	node u,v;
	int i,j;
	edge e, ee;
	INSTRUCTIONS_TYPES it;
	this->clear();
	this->set_ISA(G.get_ISA());

	// low level node insertion to keep same node ids
	forall_nodes(u, G){
		it=G.instruction_type(u);
		new_node(it.opcode_id(), G.id(u));
	}
	
	//  edge insertion 
	forall_edges(e, G){
		u=G.source(e);
		v=G.target(e);
		new_edge(G.id(u),G.id(v), G[e]);
	}
	  
	build_internal_structures();  
}

void LOOP::copy(const LOOP &G, node_map<node> &mn, edge_map<edge> &me) {
	node u,v;
	int i,j;
	edge e, ee;
	mn.init(G);
	me.init(G);
	INSTRUCTIONS_TYPES it;
	INFO_EDGE ie;
	this->clear();
	this->set_ISA(G.get_ISA());

	// low level node insertion to keep same node ids
	forall_nodes(u, G){
		it=G.instruction_type(u);
		i=this->new_node(it.opcode_id(), G.id(u));
		mn[u]=this->node_with_id(i);
	}
	
	//  edge insertion 
	forall_edges(e, G){
		u=G.source(e);
		v=G.target(e);
		ee=new_edge(G.id(u),G.id(v), G[e]);
		me[e]=ee;
	}
	  
	build_internal_structures();  
}


void LOOP::clear(){
	int i;
        cpt_nodes=0;;
        GRAPH<INFO_NODE, INFO_EDGE>::clear();
        node_index_table.clear();
        Cons.clear();
        E_R.clear();
        V_R.clear();
}



set<edge> LOOP::flow_from_to(node u, node v) const{
	edge e;
	set<edge> out_edge_u=edges_from_to(u,v);
	INFO_EDGE ie;
	forall(e, out_edge_u){
		ie=inf(e);
		if (!ie.is_flow()) out_edge_u.del(e);
	}
	return out_edge_u;
}

set<edge> LOOP::flow_from_to(int iu, int iv) const{
	edge e;
	set<edge> out_edge_u=edges_from_to(iu,iv);
	INFO_EDGE ie;
	forall(e, out_edge_u){
		ie=inf(e);
		if (!ie.is_flow()) out_edge_u.del(e);
	}
	return out_edge_u;
}

set<edge> LOOP::flow_reg_from_to(node u, node v)const{
	edge e;
	int t;
	if (ISA.number_of_registers_types()>1) {
		cerr<<"DDG error : flow_reg_from_to called while multiple registers types exist.\n";
		throw NonUniqueRegisterType();
	}
	t=ISA.get_the_unique_register_type_id();
	
	set<edge> out_edge_u= flow_from_to(u,v) ;
	forall(e, out_edge_u){
		if (!is_flow_reg(e,t)) out_edge_u.del(e);
	}
	return out_edge_u;
}

set<edge> LOOP::flow_reg_from_to(int iu, int iv) const{
	edge e;
	int t;
	if (ISA.number_of_registers_types()>1) {
		cerr<<"DDG error : flow_reg_from_to called while multiple registers types exist.\n";
		throw NonUniqueRegisterType();
	}
	t=ISA.get_the_unique_register_type_id();
	set<edge> out_edge_u= flow_from_to(iu,iv) ;
	forall(e, out_edge_u){
		if (!is_flow_reg(e,t)) out_edge_u.del(e);
	}
	return out_edge_u;
}


set<edge> LOOP::flow_reg_from_to(node u, node v, int t)const{
	edge e;
	
	
	set<edge> out_edge_u= flow_from_to(u,v) ;
	forall(e, out_edge_u){
		if (!is_flow_reg(e,t)) out_edge_u.del(e);
	}
	return out_edge_u;
}

set<edge> LOOP::flow_reg_from_to(int iu, int iv, int t) const{
	edge e;
	
	set<edge> out_edge_u= flow_from_to(iu,iv) ;
	forall(e, out_edge_u){
		if (!is_flow_reg(e,t)) out_edge_u.del(e);
	}
	return out_edge_u;
}

int LOOP::max_dist(node u, node v) const{
	edge e;
	int i=-MAXINT,j;
	set<edge> SE=edges_from_to(u,v);
	forall(e,SE){
		if((j=lambda(e))>i){
			i=j;
		}
	}
	return i;
}

int LOOP::max_flow_dist(node u, node v) const{
	edge e;
	int i=-MAXINT,j;
	set<edge> SE=flow_from_to(u,v);
	forall(e,SE){
		if((j=lambda(e))>i){
			i=j;
		}
	}
	return i;
}

int LOOP::max_flow_reg_dist(node u, node v) const{
	edge e;
	int i=-MAXINT,j;
	if (ISA.number_of_registers_types()>1) {
		cerr<<"DDG error : max flow_reg distance called while multiple registers types exist.\n";
		throw NonUniqueRegisterType();
	}
	set<edge> SE=flow_reg_from_to(u,v);
	forall(e,SE){
		if((j=lambda(e))>i){
			i=j;
		}
	}
	return i;
}

int LOOP::max_flow_reg_dist(node u, node v, int t) const{
	edge e;
	int i=-MAXINT,j;
	
	set<edge> SE=flow_reg_from_to(u,v,t);
	forall(e,SE){
		if((j=lambda(e))>i){
			i=j;
		}
	}
	return i;
}

int LOOP::max_dist(int u, int v)const{
	edge e;
	int i=-MAXINT,j;
	set<edge> SE=edges_from_to(u,v);
	forall(e,SE){
		if((j=lambda(e))>i){
			i=j;
		}
	}
	return i;
}

int LOOP::max_flow_dist(int u, int v)const {
	edge e;
	int i=-MAXINT,j;
	set<edge> SE=flow_from_to(u,v);
	forall(e,SE){
		if((j=lambda(e))>i){
			i=j;
		}
	}
	return i;
}


int LOOP::max_flow_reg_dist(int u, int v) const{
	edge e;
	int i=-MAXINT,j;
	if (ISA.number_of_registers_types()>1) {
		cerr<<"DDG error : max flow_reg distance called while multiple registers types exist.\n";
		throw NonUniqueRegisterType();
	}
	
	set<edge> SE=flow_reg_from_to(u,v);
	forall(e,SE){
		if((j=lambda(e))>i){
			i=j;
		}
	}
	return i;
}

int LOOP::max_flow_reg_dist(int u, int v, int t) const{
	edge e;
	int i=-MAXINT,j;
	
	set<edge> SE=flow_reg_from_to(u,v, t);
	forall(e,SE){
		if((j=lambda(e))>i){
			i=j;
		}
	}
	return i;
}


int LOOP::min_dist(node u, node v) const{
	edge e;
	int i=MAXINT,j;
	set<edge> SE=edges_from_to(u,v);
	forall(e,SE){
		if((j=lambda(e))<i){
			i=j;
		}
	}
	return i;
}

int LOOP::min_flow_dist(node u, node v) const{
	edge e;
	int i=MAXINT,j;
	
	set<edge> SE=flow_from_to(u,v);
	forall(e,SE){
		if((j=lambda(e))<i){
			i=j;
		}
	}
	return i;
}

int LOOP::min_flow_reg_dist(node u, node v) const{
	edge e;
	int i=MAXINT,j;
	
	if (ISA.number_of_registers_types()>1) {
		cerr<<"DDG error : min flow_reg distance called while multiple registers types exist.\n";
		throw NonUniqueRegisterType();
	}
	
	
	set<edge> SE=flow_reg_from_to(u,v);
	forall(e,SE){
		if((j=lambda(e))<i){
			i=j;
		}
	}
	return i;
}

int LOOP::min_flow_reg_dist(node u, node v, int t) const{
	edge e;
	int i=MAXINT,j;
		
	set<edge> SE=flow_reg_from_to(u,v, t);
	forall(e,SE){
		if((j=lambda(e))<i){
			i=j;
		}
	}
	return i;
}

int LOOP::min_dist(int u, int v) const{
	edge e;
	int i=MAXINT,j;
	set<edge> SE=edges_from_to(u,v);
	forall(e,SE){
		if((j=lambda(e))<i){
			i=j;
		}
	}
	return i;
}

int LOOP::min_flow_dist(int u, int v) const{
	edge e;
	int i=MAXINT,j;
	set<edge> SE=flow_from_to(u,v);
	forall(e,SE){
		if((j=lambda(e))<i){
			i=j;
		}
	}
	return i;
}

int LOOP::min_flow_reg_dist(int u, int v) const{
	edge e;
	int i=MAXINT,j;
	
	if (ISA.number_of_registers_types()>1) {
		cerr<<"DDG error : min flow_reg distance called while multiple registers types exist.\n";
		throw NonUniqueRegisterType();
	}
	
	
	set<edge> SE=flow_reg_from_to(u,v);
	forall(e,SE){
		if((j=lambda(e))<i){
			i=j;
		}
	}
	return i;
}


int LOOP::min_flow_reg_dist(int u, int v, int t) const{
	edge e;
	int i=MAXINT,j;
	
	
	set<edge> SE=flow_reg_from_to(u,v, t);
	forall(e,SE){
		if((j=lambda(e))<i){
			i=j;
		}
	}
	return i;
}


void LOOP::build_internal_structures(){
	edge e; 
	node u, v;
	INFO_EDGE ie;
	list<edge> el;
	list<edge> le;
	INSTRUCTIONS_TYPES it;
	LEDA::list<int> lr;
	int m,t;
	LEDA::rational cc; // critical cycle
	int nbnode=number_of_nodes(), i,j;
	
	/* set of values in G */
	V_R.clear();
	forall_nodes(u, *this){
		if ((instruction_type(u)).is_value()) {
			lr=(instruction_type(u)).get_list_regtypes_id();
			forall(t, lr){
				V_R[t].insert(u);
			}
			
		}
	}
	
	/*---------------Consumers ----------------------------*/
	
	Cons.clear();
	lr=ISA.get_list_regtypes_id();
	forall(t, lr){
		forall(e, E_R[t]){
			Cons(source(e),t).insert(target(e));
		}
	}

	
}

void LOOP::write_to_xml(const  char *filename) const{
  ofstream fp;

  try{
    fp.open(filename);

    fp << "<ddg version='1'>\n";

    //@todo : take care of hidden nodes/edges
    
    // First, list all nodes
    const LEDA::list<node>& nodes = all_nodes();
    node n;
    forall(n, nodes){
      std::string type = instruction_type(n).opcode();
      int my_id = id(n);
      std::string description = get_string_attribute(n); 
      fp << "<operation type='" << type <<"' id='" << my_id << "'";
      if(description.length() > 0){
        // Escape the double quotes
        std::string str2("\\");
        int position = -2;
        while((position = description.find("\"", position+2)) != std::string::npos){
          description.insert(position, str2);
        }
        fp << " description=\"" << description << "\"";
      }        
      fp << "/>\n";
    }

    // Then list all edges
    const LEDA::list<edge>& edges = all_edges();
    edge e;
    forall(e, edges){
      int src = id(source(e));
      int dst = id(target(e));
      int latency     = inf(e).delta();
      int distance    = inf(e).lambda();
      DDG::edge_type et    = inf(e).etype();
      fp << "<edge src='" << src << "' dst='" << dst << "' latency='" << latency << "' distance='" << distance << "' typedep='" << edge_type_to_string(et) << "'";
      switch(et){
          case flowdep_reg :
          case antidep_reg :
          case outputdep_reg :
          case inputdep_reg :
          case killerdep :
          case reusedep :
            std::string regtype = ISA.get_register_typename(get_register_type_id(e));
            fp << " regtype='" << regtype << "'";
      }
      fp << "/>\n";
    }
    fp << "</ddg>\n";
    fp.close();
  }catch( ... ){
    cerr << "DDG error encountered while writing DAG to xml\n";
  }
}

bool LOOP::read_from_xml(const char *xml_file)
{
  int result = -1;
  
  try{
    XMLPlatformUtils::Initialize();
  } catch ( ... ){
    cerr << "DDG error while initializing XML " << xml_file<< endl;
    return false;
  }
  this -> clear();
  xercesc::XercesDOMParser *config_parser = new XercesDOMParser;
  // Tags and attributes we are interested in parsing.
  XMLCh* TAG_DDG        = XMLString::transcode("ddg");
  XMLCh* TAG_Operation  = XMLString::transcode("operation");
  XMLCh* TAG_Edge       = XMLString::transcode("edge");
  
  XMLCh* ATTR_Version     = XMLString::transcode("version");
  XMLCh* ATTR_Distance    = XMLString::transcode("distance");
  XMLCh* ATTR_Latency     = XMLString::transcode("latency");
  XMLCh* ATTR_Regtype     = XMLString::transcode("regtype");
  XMLCh* ATTR_Src         = XMLString::transcode("src");
  XMLCh* ATTR_Dst         = XMLString::transcode("dst");
  XMLCh* ATTR_Type        = XMLString::transcode("type");
  XMLCh* ATTR_Typedep     = XMLString::transcode("typedep");
  XMLCh* ATTR_Description = XMLString::transcode("description");
  XMLCh* ATTR_Id          = XMLString::transcode("id");
  
  try{
    // Check if file exists and can be read
    struct stat filebuf;
    int status;
    status = stat(xml_file, &filebuf);
    
    if (status != 0){
      cerr << "DDG error. Cannot stat XML file " << xml_file<< endl;
      InvalidISADesc inv;
      throw (inv);
    }
    
    // @todo : use DTD
    config_parser->setValidationScheme( XercesDOMParser::Val_Auto);
    
    config_parser->parse(xml_file);
    
    DOMDocument* xml_doc = config_parser->getDocument();
    
    DOMElement* root = xml_doc->getDocumentElement();
    if( !root ){
      cerr << "DDG error with input XML file" << xml_file<< endl;
      InvalidISADesc inv;
      throw (inv);
    }
    
    int version  = XMLString::parseInt(root->getAttribute(ATTR_Version));
    
    // @todo Only support version 1
    if(version != 1){
      cerr << "DDG error. Wrong XML version " << xml_file<< endl;
      InvalidISADesc inv;
      throw (inv);
    }

    // First, operations 
    DOMNodeList* operations = root->getElementsByTagName(TAG_Operation);
    XMLSize_t node_count = operations->getLength();
    
    for( XMLSize_t x = 0; x < node_count; ++x ) {
      DOMNode* current_node = operations->item(x);
      if( current_node->getNodeType() && current_node->getNodeType() == DOMNode::ELEMENT_NODE ){
        DOMElement* current_element = dynamic_cast< xercesc::DOMElement* >( current_node );
        
        // Create the new operation 
        char *type = XMLString::transcode(current_element->getAttribute(ATTR_Type));
        int my_id     = XMLString::parseInt(current_element->getAttribute(ATTR_Id));
        char *description = XMLString::transcode(current_element->getAttribute(ATTR_Description));
        std::string stype = std::string(type);
        if(!ISA.exist_instruction_type_opcode(stype)){
          cerr <<"DDG Error.  Opcode non recognized: "<< stype<< " in "<< xml_file<<endl;
          BadIODDG bioddg;
          throw bioddg;
        }
        new_node(ISA.get_instruction_type_by_opcode(stype).opcode_id(), my_id, std::string(description));
        XMLString::release(&description);
        XMLString::release(&type);
      }
    }
    
    DOMNodeList* edges = root->getElementsByTagName(TAG_Edge);
    node_count = edges->getLength();
    
    for( XMLSize_t x = 0; x < node_count; ++x ) {
      DOMNode* current_node = edges->item(x);
      if( current_node->getNodeType() && current_node->getNodeType() == DOMNode::ELEMENT_NODE ){
        DOMElement* current_element = dynamic_cast< xercesc::DOMElement* >( current_node );
        
        // Create the new edge 
        int src      = XMLString::parseInt(current_element->getAttribute(ATTR_Src));
        int dst      = XMLString::parseInt(current_element->getAttribute(ATTR_Dst));
        char *typedep  = XMLString::transcode(current_element->getAttribute(ATTR_Typedep));
        int latency  = XMLString::parseInt(current_element->getAttribute(ATTR_Latency));
        int distance = XMLString::parseInt(current_element->getAttribute(ATTR_Distance));
       
        node s = node_with_id(src);
        node t = node_with_id(dst);
        DDG::edge_type et = string_to_edge_type(std::string(typedep));
        INFO_EDGE ie = INFO_EDGE(latency, distance, et);
        switch(et){
          case flowdep_reg :
          case antidep_reg :
          case outputdep_reg :
          case inputdep_reg :
          case killerdep :
          case reusedep :
            char *reg_type = XMLString::transcode(current_element->getAttribute(ATTR_Regtype));
            int id = ISA.get_register_type_id(std::string(reg_type));
            if(id == -1){
              cerr << "DDG error. Bad register type" << std::string(reg_type) << "in " << xml_file<<endl;
              throw InvalidISADesc();
            }
            ie.set_register_type_id(id);
            XMLString::release(&reg_type);
        }
        XMLString::release(&typedep);
        new_edge(s, t, ie);
      }
    }
    result = 1;
    build_internal_structures();
    
  }catch( ... ){
    cerr<<"DDG error while parsing XML file "<<xml_file<<endl;
    result = -1;
  }
  
  // Clean up
  XMLString::release(&TAG_DDG       );
  XMLString::release(&TAG_Operation );
  XMLString::release(&TAG_Edge      );
  XMLString::release(&ATTR_Version  );
  XMLString::release(&ATTR_Distance );
  XMLString::release(&ATTR_Latency  );
  XMLString::release(&ATTR_Regtype  );
  XMLString::release(&ATTR_Src      );
  XMLString::release(&ATTR_Dst      );
  XMLString::release(&ATTR_Type     );
  XMLString::release(&ATTR_Typedep  );
  XMLString::release(&ATTR_Description);
  XMLString::release(&ATTR_Id);
  XMLPlatformUtils::Terminate();
  
  return true;
}

void LOOP::write_to_gl(const char * file_name) const{
	node s, t;
	edge e;
	int nbnode, nbedge, i;
	INSTRUCTIONS_TYPES it;
	ofstream gfile; /* file.g and file.l */
	gfile.open(file_name);
	if(!gfile.is_open()){
		cerr<<"DDG IO failure : could not open " << file_name<<endl;
		return;
	}
	if(ISA.number_of_registers_types()>1){
		cerr<<"DDG warning : writing to gl file while the number of register types is greater than 1.\n";
	}
	
	gfile<<"DDG.GRAPH\n";
	nbnode=number_of_nodes();
	gfile<<nbnode<<endl;
	for(i=1;i<=nbnode;i++){
		try{
			s=node_with_id(i);
		}
		catch(IllegalAccessNode){
			continue;
		}
		it=instruction_type(s);
		gfile<<it.opcode_id()<<endl;
	}
	nbedge=number_of_edges();
	gfile  <<nbedge<<endl;
		
	forall_edges(e,*this){
		gfile<<id(source(e))<<'\t'<<id(target(e))<<'\t'<<delta(e)<<'\t'<< lambda (e) <<'\t'<< edge_type_to_string(etype(e)) <<endl; 
	}
	
	gfile.close(); 
}

void LOOP::write_to_gl(const char * file_name, int regtype_id) const{
	node s, t;
	edge e;
	int nbnode, nbedge, i;
	INSTRUCTIONS_TYPES it;
	ofstream gfile; /* file.g and file.l */
	gfile.open(file_name);
	if(!gfile.is_open()){
		cerr<<"DDG IO failure : could not open " << file_name<<endl;
		return;
	}
	if(ISA.number_of_registers_types()>1){
		cerr<<"DDG warning : writing to gl file while the number of register types is greater than 1.\n";
	}
	
	gfile<<"DDG.GRAPH\n";
	nbnode=number_of_nodes();
	gfile<<nbnode<<endl;
	for(i=1;i<=nbnode;i++){
		try{
			s=node_with_id(i);
		}
		catch(IllegalAccessNode){
			continue;
		}
		it=instruction_type(s);
		gfile<<it.opcode_id()<<endl;
	}
	nbedge=number_of_edges();
	gfile  <<nbedge<<endl;
		
	forall_edges(e,*this){
		if (is_flow_reg(e,regtype_id)){
			gfile<<id(source(e))<<'\t'<<id(target(e))<<'\t'<<delta(e)<<'\t'<<0<<'\t'<< edge_type_to_string(etype(e)) <<endl; // lambda(e)=0 for any edge in a DAG
		}
		else {
			gfile<<id(source(e))<<'\t'<<id(target(e))<<'\t'<<delta(e)<<'\t'<< lambda (e) <<'\t'<< edge_type_to_string(serial) <<endl; 
		}
	}
	
	gfile.close(); 
}

bool LOOP::read_from_gl(const char * file_name){
	node s, t;
	edge e;
	int nbnode, nbedge, i, sourc, targe, delta, lambda=0;
	int tmp;
	DDG::edge_type et;
	std::string bidon;
	INFO_EDGE ie;
	INFO_NODE in;
	char cc[MAX_STRING];
	INSTRUCTIONS_TYPES it;
	list_item li;
	ifstream gfile; /* file.g and file.l */
	gfile.open(file_name);
	clear();
	if(!gfile.is_open()){
		cerr<<"DDG IO Failure: could not open " << file_name<<endl;
		BadIODDG bioddg;
		throw bioddg;
	}
	gfile.getline(cc,MAX_STRING,'\n'); // get tje header section
	bidon=std::string(cc);
	
	std::string str_etype;
	i=bidon.find("DDG.GRAPH"); // looks for the gl header section pecifier
	if ((i<0)||i>bidon.size()) {
		cerr <<"DDG error : bad gl header section : "<<file_name<<". Seems not a valid gl  format (version 1.1).\n";
			gfile.close();
			BadIODDG bioddg;
			throw bioddg;
	}
	gfile>>nbnode;
	for(i=0;i<nbnode;i++){
		gfile>>tmp; /* read the opcode*/
		/* search the opcode in the set of instructions types */
		
		if(!ISA.exist_instruction_type_id(tmp)){
			cerr <<"DDG Error.  Opcode non recognized: "<<tmp<<endl;
			BadIODDG bioddg;
			throw bioddg;
		}
		else {
			new_node(tmp);
		}
	}
	gfile  >>nbedge;
	/* an edge is defined by : source target delta*/
	for(i=0;i<nbedge;i++){
		gfile>>sourc>>targe>>delta>>lambda>>str_etype;
		try{
			s=node_with_id(sourc);
			t=node_with_id(targe);
		}
		catch(IllegalAccessNode){
			continue;
		}
		try{
			et=string_to_edge_type(str_etype);
		}
		catch(IllegalDependance){
			continue;
		}
		ie=INFO_EDGE(delta, lambda, et);
		if (ie.is_flow_reg()){
			if (ISA.number_of_registers_types()!=1){
				cerr<<"DDG Warning : read from gl while there is not a unique register type in the ISA.\n";
				throw NonUniqueRegisterType();
			}
			ie.set_register_type_id(ISA.get_the_unique_register_type_id());
			
		}
		e=new_edge(s, t, ie);
		
	}
	
	gfile.close();
	build_internal_structures();
	return true;
}


bool LOOP::read_from_gl(const char * file_name, int regtype_id){
	node s, t;
	edge e;
	int nbnode, nbedge, i, sourc, targe, delta, lambda=0;
	int tmp;
	DDG::edge_type et;
	std::string bidon;
	INFO_EDGE ie;
	INFO_NODE in;
	char cc[MAX_STRING];
	INSTRUCTIONS_TYPES it;
	list_item li;
	ifstream gfile; /* file.g and file.l */
	gfile.open(file_name);
	clear();
	if(!gfile.is_open()){
		cerr<<"DDG IO Failure: could not open " << file_name<<endl;
		BadIODDG bioddg;
		throw bioddg;
	}
	gfile.getline(cc,MAX_STRING,'\n'); // get tje header section
	bidon=std::string(cc);
	
	std::string str_etype;
	i=bidon.find("DDG.GRAPH"); // looks for the gl header section pecifier
	if ((i<0)||i>bidon.size()) {
		cerr <<"DDG error : bad gl header section : "<<file_name<<". Seems not a valid gl  format (version 1.1).\n";
			gfile.close();
			BadIODDG bioddg;
			throw bioddg;
	}
	gfile>>nbnode;
	for(i=0;i<nbnode;i++){
		gfile>>tmp; /* read the opcode*/
		/* search the opcode in the set of instructions types */
		
		if(!ISA.exist_instruction_type_id(tmp)){
			cerr <<"DDG Error.  Opcode non recognized: "<<tmp<<endl;
			BadIODDG bioddg;
			throw bioddg;
		}
		else {
			new_node(tmp);
		}
	}
	gfile  >>nbedge;
	/* an edge is defined by : source target delta*/
	for(i=0;i<nbedge;i++){
		gfile>>sourc>>targe>>delta>>lambda>>str_etype;
		try{
			s=node_with_id(sourc);
			t=node_with_id(targe);
		}
		catch(IllegalAccessNode){
			continue;
		}
		try{
			et=string_to_edge_type(str_etype);
		}
		catch(IllegalDependance){
			continue;
		}
		ie=INFO_EDGE(delta, lambda, et);
		if (ie.is_flow_reg()){
			ie.set_register_type_id(regtype_id);
			
		}
		e=new_edge(s, t, ie);		
	}
	
	gfile.close();
	build_internal_structures();
	return true;
}

void LOOP::write_to_vcg(const char *filename) const
{
	ofstream ofile(filename);
	node v, src, dest;
	edge e;
	INSTRUCTIONS_TYPES it;
	INFO_EDGE ie;
	INFO_NODE in;
	set<node> SN;
	REGISTER_TYPES t;
	LEDA::list<REGISTER_TYPES> T=ISA.T();
	if(!ofile){
	cerr<< "DDG IO failure: could not create "<<filename<<'\n';
		return;
	}

	
	
	ofile<<"graph: {\ntitle: \""<<filename<<"\"\nnode.shape : ellipse\ninfoname1 : \"ISA opcode\"\nsplines : yes\nedge.color : black\nlayout_downfactor:2\nlayout_upfactor:1\nlayoutalgorithm : maxdepth\ndisplay_edge_labels : yes\n";
	ofile<<"infoname2 : \"Latencies\"\n";
	ofile<<"infoname3 : \"Textual Attribute\"\n";
	
	ofile<<"classname "<< flowdep_reg<<" : \"" <<edge_type_to_string(flowdep_reg)<< " \"\n";
	ofile<<"classname "<< antidep_reg<<" : \"" <<edge_type_to_string(antidep_reg)<< " \"\n";
	ofile<<"classname "<< outputdep_reg<<" : \"" <<edge_type_to_string(outputdep_reg)<< " \"\n";
	ofile<<"classname "<< inputdep_reg<<" : \"" <<edge_type_to_string(inputdep_reg)<< " \"\n";
	ofile<<"classname "<< flowdep_mem<<" : \"" <<edge_type_to_string(flowdep_mem)<< " \"\n";
	
	ofile<<"classname "<< antidep_mem<<" : \"" <<edge_type_to_string(antidep_mem)<< " \"\n";
	ofile<<"classname "<< outputdep_mem<<" : \"" <<edge_type_to_string(outputdep_mem)<< " \"\n";
	ofile<<"classname "<< inputdep_mem<<" : \"" <<edge_type_to_string(inputdep_mem)<< " \"\n";
	ofile<<"classname "<< spilldep_mem<<" : \"" <<edge_type_to_string(spilldep_mem)<< " \"\n";
	ofile<<"classname "<< other_mem<<" : \"" <<edge_type_to_string(other_mem)<< " \"\n";
	
	ofile<<"classname "<< serial<<" : \"" <<edge_type_to_string(serial)<< " \"\n";
	ofile<<"classname "<< killerdep<<" : \"" <<edge_type_to_string(killerdep)<< " \"\n";
	ofile<<"classname "<< reusedep<<" : \"" <<edge_type_to_string(reusedep)<< " \"\n";
	
	
	
	forall_nodes(v, *this){
		bool isvalue=false;
		it=this->instruction_type(v);
		ofile<<"node: { title: \""<<id(v)<< "\"  info1: \""<<it.opcode()<<"\" info2: \" writes " ;
		forall(t, T){
			 if (this->is_value(v,t.get_register_type_id())){
			 	isvalue=true;
			 	ofile<<t.get_register_typename()<<"("<<t.get_register_type_id()<<") ";
		 }	
		}
		if(! isvalue){
			ofile <<". lat="<<latency(v)<<", dr="<<delta_r(v);
		}
		ofile <<"\" info3: \""<<get_string_attribute(v)<< "\" color:";
		if(isvalue){
			ofile<<"green}\n";
		}
		else ofile<<"lightgrey}\n";
				
	}
  // unparsing edges in vcg format
	forall_edges(e, *this){
		src = source(e);
		dest = target(e);
		ie=this->inf(e);
		ofile<<"edge: { sourcename: \"";
		ofile<<id(src);
		
		
		ofile<<"\" targetname: \"";
		
		ofile<<id(dest);

		
		
		if (ie.get_register_type_id()<0)
			ofile<<"\" class:"<<etype(e)<<" label:\"("<< delta(e)<<" , "<<lambda(e) <<")\" color:";
		else
			ofile<<"\" class:"<<etype(e)<<" label:\"("<< delta(e)<<" , "<<lambda(e) << " , " << this->ISA.get_register_typename(ie.get_register_type_id()) <<")\" color:";
			
		switch (etype(e)){
		case flowdep_reg: 
		case flowdep_mem: ofile<< "red}\n";
			break;		
		case antidep_reg: 
		case antidep_mem: ofile<<"green}\n";
			break;
		case outputdep_reg: 
		case outputdep_mem: ofile<<"darkmagenta}\n";
			break;	
		case inputdep_reg:
		 	inputdep_mem :ofile<<"blue}\n";
			break;
		default: serial: ofile<<"black}\n";
		}	
		
		
	}// end for edge
	ofile<<"}\n"; 
	
}

void LOOP::write_to_vcg(const char *filename, int t) const
{
	ofstream ofile(filename);
	node v, src, dest;
	edge e;
	INSTRUCTIONS_TYPES it;
	INFO_EDGE ie;
	INFO_NODE in;
	set<node> SN;
	if(!ofile){
		cerr<< "DDG IO failure: could not create "<<filename<<'\n';
		return;
	}

	
	
	ofile<<"graph: {\ntitle: \""<<filename<<"\"\nnode.shape : ellipse\ninfoname1 : \"ISA opcode\"\nsplines : yes\nedge.color : black\nlayout_downfactor:2\nlayout_upfactor:1\nlayoutalgorithm : maxdepth\ndisplay_edge_labels : yes\n";
	ofile<<"infoname2 : \"Latencies\"\n";
	ofile<<"infoname3 : \"Textual Attribute\"\n";
	
	ofile<<"classname "<< flowdep_reg<<" : \"" <<edge_type_to_string(flowdep_reg)<< " \"\n";
	ofile<<"classname "<< antidep_reg<<" : \"" <<edge_type_to_string(antidep_reg)<< " \"\n";
	ofile<<"classname "<< outputdep_reg<<" : \"" <<edge_type_to_string(outputdep_reg)<< " \"\n";
	ofile<<"classname "<< inputdep_reg<<" : \"" <<edge_type_to_string(inputdep_reg)<< " \"\n";
	ofile<<"classname "<< flowdep_mem<<" : \"" <<edge_type_to_string(flowdep_mem)<< " \"\n";
	
	ofile<<"classname "<< antidep_mem<<" : \"" <<edge_type_to_string(antidep_mem)<< " \"\n";
	ofile<<"classname "<< outputdep_mem<<" : \"" <<edge_type_to_string(outputdep_mem)<< " \"\n";
	ofile<<"classname "<< inputdep_mem<<" : \"" <<edge_type_to_string(inputdep_mem)<< " \"\n";
	ofile<<"classname "<< spilldep_mem<<" : \"" <<edge_type_to_string(spilldep_mem)<< " \"\n";
	ofile<<"classname "<< other_mem<<" : \"" <<edge_type_to_string(other_mem)<< " \"\n";
	
	ofile<<"classname "<< serial<<" : \"" <<edge_type_to_string(serial)<< " \"\n";
	ofile<<"classname "<< killerdep<<" : \"" <<edge_type_to_string(killerdep)<< " \"\n";
	ofile<<"classname "<< reusedep<<" : \"" <<edge_type_to_string(reusedep)<< " \"\n";
	
	
	
	forall_nodes(v, *this){
		bool isvalue=false;
		it=this->instruction_type(v);
		ofile<<"node: { title: \""<<id(v)<< "\"  info1: \""<<it.opcode()<<"\" info2: \" writes " ;
		if (this->is_value(v,t)){
				isvalue=true;
				ofile<<ISA.get_register_typename(t)<<"(regtype_id="<<t<<") ";
		}	
		
		if(! isvalue){
			ofile <<". lat="<<latency(v)<<", dr="<<delta_r(v);
		}
		ofile <<"\" info3: \""<<get_string_attribute(v)<< "\" color:";
		if(isvalue){
			ofile<<"green}\n";
		}
		else ofile<<"lightgrey}\n";
				
	}
  // unparsing edges in vcg format
	forall_edges(e, *this){
		src = source(e);
		dest = target(e);
		ie=this->inf(e);
		ofile<<"edge: { sourcename: \"";
		ofile<<id(src);
		
		
		ofile<<"\" targetname: \"";
		
		ofile<<id(dest);

		
		
		if (ie.get_register_type_id()<0)
			ofile<<"\" class:"<<etype(e)<<" label:\"("<< delta(e)<<" , "<<lambda(e) <<")\" color:";
		else
			ofile<<"\" class:"<<etype(e)<<" label:\"("<< delta(e)<<" , "<<lambda(e) << " , " << this->ISA.get_register_typename(ie.get_register_type_id()) <<")\" color:";
			
		switch (etype(e)){
			case flowdep_reg: 
				case flowdep_mem: ofile<< "red}\n";
				break;		
			case antidep_reg: 
				case antidep_mem: ofile<<"green}\n";
				break;
			case outputdep_reg: 
				case outputdep_mem: ofile<<"darkmagenta}\n";
				break;	
			case inputdep_reg:
				inputdep_mem :ofile<<"blue}\n";
				break;
				default: serial: ofile<<"black}\n";
		}	
		
		
	}// end for edge
	ofile<<"}\n"; 
	
}



// loop merge
void loop_merge(const LOOP &L1, const LOOP &L2, LOOP &L3){
	node u, v;
	edge e;
	L3.clear();
	L3.set_ISA(L1.get_ISA());
	node_array<int> node_id1(L1), node_id2(L2) ;

	/* add the nodes of L1 into L3 */
	forall_nodes(u,L1){
		node_id1[u]=L3.new_node((L1.instruction_type(u)).opcode_id());
	}
	/* add the adges of L1 into L3 */
	forall_edges(e, L1){
		L3.new_edge( node_id1[L1.source(e)], node_id1[L1.target(e)], L1[e] );
	}

	/* add the nodes of L2 into L3 */
	forall_nodes(u,L2){
		node_id2[u]=L3.new_node((L2.instruction_type(u)).opcode_id());
	}
	/* add the adges of L2 into L3 */
	forall_edges(e, L2){
		L3.new_edge( node_id2[L2.source(e)], node_id2[L2.target(e)], L2[e] );
	}
	L3.build_internal_structures();
}

/* build a DAG from a loop unroled j times */

void unroll(const LOOP& L, DAG &unrolled, int j){
	node u, v, ui, vi;
	edge e;
	map2<node,int, int> node_copy;
	INFO_EDGE ie, iee;
	INFO_NODE in;
	int iter; /* couter */
	unrolled.clear();
	unrolled.set_ISA(L.get_ISA());
	if (j<0) return; //do nothing if the unrolling degree is negative
	forall_edges(e, L)
	{ if (L.lambda(e)<0) return; } //all distances should be nonnegative, otherwie do nothing
	//create j copies of for each node
	for(iter=0;iter<j+1;iter++){
		forall_nodes(u, L){
			in=L[u];
			node_copy(u,iter)=(unrolled.new_node(in.instruction_type_id()));
		}
	}
	
	//insert the edges
	forall_edges(e, L){
		ie=L[e];
		u=L.source(e);
		v=L.target(e);
		// and edger e=(u,v) in the original loop has to be unrolled j times. 
		//If u_i is the i^th copy of the node u, then the edge e=(u,v) in the original loop becomes an edge e' from u_i to v_{(i+\lambda(e)) mod j} 
		// with a latency \delta (e')=\delta(e) and a distance \lambda(e')=(i+\lambda(e)) div j
		for(iter=0;iter<j+1;iter++){
			ui=(unrolled.node_with_id(node_copy(u,iter)));
			vi=(unrolled.node_with_id(node_copy(v,  (iter+ ie.lambda()) % (j+1) ) ) ) ; 
			iee=ie; // set the same edge latency and type
			iee.set_lambda((iter+ie.lambda())/(j+1)); // correct the distance
			if (iee.lambda()==0) { // edge in the body of the unrolled loop => insert it
				unrolled.new_edge(ui, vi, iee);
			}
		}
	}
	
	unrolled.build_internal_structures();

}

/* returns a loop unroled j times */

void unroll(const LOOP& L, LOOP &unrolled, int j){
	node u, v, ui, vi;
	edge e;
	map2<node,int, int> node_copy;
	INFO_EDGE ie, iee;
	INFO_NODE in;
	int iter; /* couter */
	unrolled.clear();
	unrolled.set_ISA(L.get_ISA());
	if (j<0) return; //do nothing if the unrolling degree is negative
	forall_edges(e, L)
	{ if (L.lambda(e)<0) return; } //all distances should be nonnegative, otherwise do nothing
		
	//create j copies of for each node
	for(iter=0;iter<j+1;iter++){
		forall_nodes(u, L){
			in=L[u];
			node_copy(u,iter)=(unrolled.new_node((L.instruction_type(u)).opcode_id()));
		}
	}
	
	//insert the edges
	forall_edges(e, L){
		ie=L[e];
		u=L.source(e);
		v=L.target(e);
		// and edger e=(u,v) in the original loop has to be unrolled j times. 
		//If u_i is the i^th copy of the node u, then the edge e=(u,v) in the original loop becomes an edge e' from u_i to v_{(i+\lambda(e)) mod j} 
		// with a latency \delta (e')=\delta(e) and a distance \lambda(e')=(i+\lambda(e)) div j
		for(iter=0;iter<j+1;iter++){
			ui=(unrolled.node_with_id(node_copy(u,iter)));
			vi=(unrolled.node_with_id(node_copy(v,  (iter+ ie.lambda()) % (j+1) ) ) ) ; 
			iee=ie; // set the same edge latency and type
			iee.set_lambda((iter+ie.lambda())/(j+1)); // correct the distance
			unrolled.new_edge(ui, vi, iee);
		}
	}
	
	unrolled.build_internal_structures();
	
}
// Critical cycle computation
// Originally developped by Antoine Sawaya in 1991 improved by Sid Touati in 2001
//

bool LOOP::FLOYD(GRAPH<INFO_NODE, INFO_EDGE>&G, node_matrix<double>& DIST) const
{
	node_matrix<double> DIST1(G,MAXDOUBLE);
	INFO_EDGE ie;
	node v,w,x;
	edge e;
	long int i;
	double k;
  
	forall_edges(e,G)
		DIST1(G.source(e),G.target(e)) = (ie=G.inf(e)).sigma();
                                            
	x = G.first_node();
  
	for(i = 1; i <= (G.number_of_nodes()); i++)
	{
		forall_nodes(v,G)
		{
			forall_nodes(w,G)
			{
	      //infini + x  == x + infini == infini
				if ((DIST1(v,x) == MAXDOUBLE)||(DIST1(x,w) == MAXDOUBLE)) k = MAXDOUBLE;
				else
					k = DIST1(v,x) + DIST1(x,w);
				DIST(v,w) = Min(DIST1(v,w),k);
			}
			if (DIST(v,v) < 0) return false;  //detection d'un cycle negatif.
		}

		DIST1 = DIST;
		x = G.succ_node(x);
	}
	return true;
 
}


/******************************************************************************
*                                                                              *
*  MIN_CYCLE                                                                   *	
*                                                                              *
*  Fonction  prenant en parametre un graphe monovalue G; utilisant la fonction *
*  FLOYD, elle retourne :MAXDOUBLE si graphe acyclique                         *	
*                        -1 si detection d'un cycle negatif                    *
*                        valeur du cycle minimal du graphe sinon.              *
*                                                                              *
*******************************************************************************/	

double  LOOP::MIN_CYCLE(GRAPH<INFO_NODE, INFO_EDGE>& G) const
{
	double min_cycle;
	bool b;
	node v;
	node_matrix<double> DISTMIN(G,MAXDOUBLE);

	b = FLOYD (G , DISTMIN);
  
	if (b == true)
	{
		min_cycle = DISTMIN(G.first_node(),G.first_node());

		forall_nodes(v,G)
		{
			min_cycle = ((DISTMIN(v,v) < min_cycle)? DISTMIN(v,v):min_cycle);
		}

		return min_cycle;
	}
	else return -1;   //existence d'un cycle negatif.

}


/***************************************************************************
*                                                                           *
* PREPARE_CRITICAL_CYCLE                                                    *
*                                                                           *
* Fonction prenant en parametre un graphe valu� par delta et une liste      * 
* d'entiers lambda correspondant a la double valuation du graphe.           * 
* Elle determine le graphe G1 qui contient un seul circuit negatif          * 
* (ou plusieurs de valeur �gale) correspondant au circuit critique          * 
* du graphe G. Elle retourne MAXDOUBLE si graphe acyclique,                 *
* la valeur approximative du cycle critique sinon.                          *
* Le cycle critique etant: Max{delta(C)/lambda(C)} quelque soit C un cycle  *
* du graphe G.                                                              *
*                                                                           *
****************************************************************************/

double LOOP::PREPARE_CRITICAL_CYCLE(GRAPH<INFO_NODE, INFO_EDGE>& G1, const LOOP& G)   const
{
	int d;
	double inf,sup,min;
	
	double critical_cycle,new_val,epsilon;
	edge e;
	node v,w;
	INFO_EDGE u, ie;
	double deltamax=0;
	node_matrix<double>  new_label;
	bool fixed_point=false; /* = true in case of circuit C with delta(C)=0 and lambda(C)>0
	Mechanism added by Sid Touati */
	double inf_fixed_point=0, sup_fixed_point=0, min_fixed_point=0; /*used for fixed point detection
	Mechanism added by Sid Touati
									*/
  	
	forall_edges(e,G)
		if ((ie=G.inf(e)).sigma() > deltamax)  deltamax=(ie=G.inf(e)).sigma();
	epsilon = 1/(2*G.number_of_nodes()*G.number_of_nodes()*deltamax*deltamax);
  
	inf = 0;
	sup = G.number_of_nodes()*(G.MaxLambda());

  
	new_label.init(G1,MAXDOUBLE);

	critical_cycle = (inf+sup)/2;

	forall_edges(e,G1) 
	{
		new_val = ((ie=G1.inf(e)).lambda()-((ie=G1.inf(e)).sigma()*critical_cycle));
		if (new_val < new_label(G1.source(e), G1.target(e))) new_label(G1.source(e), G1.target(e)) = new_val;
	}

	forall_edges(e,G1) 
	{
		u=G1.inf(e);
		u.set_sigma(new_label(G1.source(e), G1.target(e)));
		G1.assign(e,u);
	}

  


	min = MIN_CYCLE(G1);
	if (min == MAXDOUBLE) return MAXDOUBLE;
	if (min == -1) sup = critical_cycle;
	else
		if (min > 0) inf = critical_cycle;

	while((min < 0)||(min > epsilon))
	{
		// Fixed point detection : detect a circuit C with delta(C)=0 and lambda(C)>1
		// Mechanism added by Sid Touati
		min_fixed_point=min;
		inf_fixed_point=inf;
		sup_fixed_point=sup;
		//if (inf==sup) break;

		G1.clear();
	
		G1 = G;
	
		critical_cycle = (inf+sup)/2;
		new_label.init(G1,MAXDOUBLE);


		forall_edges(e,G1) 
		{
			new_val = (ie=G1.inf(e)).lambda()-((ie=G1.inf(e)).sigma()*critical_cycle);
			if (new_val < new_label(G1.source(e), G1.target(e))) new_label(G1.source(e), G1.target(e)) = new_val;
		}
	
		forall_edges(e,G1) 
		{
			u = G1.inf(e);
			u.set_sigma(new_label(G1.source(e), G1.target(e)));
			G1.assign(e,u);
		}


		min = MIN_CYCLE(G1);
		if (min == MAXDOUBLE) return MAXDOUBLE;
		if (min == -1) sup = critical_cycle;
		else
			if (min>0) inf = critical_cycle;
		// Fixed point detection : detect a circuit C with delta(C)=0 and lambda(C)>1
		// Mechanism added by Sid Touati. Does not retrieve the circuit but just detect its existence
		
		if((min==min_fixed_point)&&(sup==sup_fixed_point)&&(inf==inf_fixed_point)&&(fixed_point)){
			break;
		}
		if((min==min_fixed_point)&&(sup==sup_fixed_point)&&(inf==inf_fixed_point)){
			fixed_point=true;
		}

	}
  
  //+++++++
	G1.clear();
  
	G1 = G;

	if ( min != 0 ) critical_cycle = sup;
  
	forall_edges(e,G1) 
	{
		u = G1.inf(e);
		u.set_sigma((ie=G1.inf(e)).lambda()-((ie=G1.inf(e)).sigma()*critical_cycle));
		G1.assign(e,u);
	}
  
	return 1/critical_cycle;
  
}







// /******************************************************************************
// *                                                                              *
// *  NEGATVE_CYCLE                                                               *       
// *                                                                              *
// *  Fonction qui prend en parametre un graphe monovalue G ayant, au maximum,    *
// *  un seul circuit elementaire negatif (ou plusieurs de valeur egale).         *
// *  Elle determine dans                                                         *
// *  negcycle_arcs la liste des arcs de ce circit negatif.                       *
// *  Elle rend la valeur du circuit negatif s'il existe et 0 sinon.   O(n3)      *  
// *  Algorithme concu par extension de  celui de floyd_warshall.                 * 
// *                                                                              *      
// *******************************************************************************/


double LOOP::NEGATVE_CYCLE(GRAPH<INFO_NODE, INFO_EDGE>& G, list<edge>& negcycle_arcs) const
{


	node_matrix<double> DIST(G,MAXDOUBLE);
	node_matrix<double> DIST1(G,MAXDOUBLE);
	node_matrix<list<edge>*> PATHMIN(G);
	node_matrix<list<edge>*> PATHMIN1(G);

	node v,w,x;
	edge e;
	INFO_EDGE ie;
	long int i;
	double k;

	negcycle_arcs.clear();
  
	forall_nodes(v,G)
	{
		forall_nodes(w,G)
		{
			PATHMIN(v,w)=new list<edge>;
			(*PATHMIN(v,w)).clear();
			PATHMIN1(v,w)=new list<edge>;
		}
	}

	forall_edges(e,G)
	{
		if ((G.source(e)==G.target(e))  && ((ie=G.inf(e)).sigma() < 0))
		{
			negcycle_arcs.append(e);
			return (ie=G.inf(e)).sigma();
		}
		if ((ie=G.inf(e)).sigma() < DIST(G.source(e),G.target(e))) 
		{DIST(G.source(e),G.target(e)) = (ie=G.inf(e)).sigma();
		DIST1(G.source(e),G.target(e)) = DIST(G.source(e),G.target(e));
		(*PATHMIN(G.source(e),G.target(e))).clear();
		(*PATHMIN(G.source(e),G.target(e))).append(e);
		PATHMIN1(G.source(e),G.target(e))=PATHMIN(G.source(e),G.target(e));
		}
	}  

	x = G.first_node();


	for(i = 1; i <= (G.number_of_nodes()); i++)
	{
		forall_nodes(v,G)
		{
			forall_nodes(w,G)
			{

				if ((DIST1(v,x) == MAXDOUBLE)||(DIST1(x,w) == MAXDOUBLE)) k = MAXDOUBLE;
				else
					k = DIST1(v,x) + DIST1(x,w);

				if (k < DIST1(v,w))
				{
					DIST(v,w) = k;
					(*PATHMIN(v,w)).clear();
		  
					forall(e,*PATHMIN1(v,x))
						(*PATHMIN(v,w)).append(e);
		    
					forall(e,*PATHMIN1(x,w))
						(*PATHMIN(v,w)).append(e);
		
				}
				else
				{
					DIST(v,w) = DIST1(v,w);
				}
			}
			if (DIST(v,v) <= 0)  //detection du circuit negatif.
			{
				forall(e,*PATHMIN(v,v))
				{
					negcycle_arcs.append(e);
				}

				return DIST(v,v); 
			}
		}
		DIST1 = DIST;
		PATHMIN1=PATHMIN;
		x = G.succ_node(x);
	}



	return 0;
 
}


// this function return the value of the critical cycle= Max[ sum(delta)/sum(lambda)]
// the parameter cc will contains the nodes of the critical cycle
LEDA::rational LOOP::CRITICAL_CYCLE(list<edge> &el) const
{
	int delta_sum = 0; 
	int lambda_sum = 0;
	INFO_EDGE ie;
	INFO_NODE in;
	int d;
	edge e1, e2, e;
	el.clear(); // empty list of edges

	GRAPH<INFO_NODE, INFO_EDGE> G1=*this;
  
	if (PREPARE_CRITICAL_CYCLE(G1, *this) == MAXDOUBLE){
		/*NoCriticalCycle nc;
		throw nc;*/
		return -1;
	}	
 	 
	list<edge> negcycle_arcs;
	NEGATVE_CYCLE(G1,negcycle_arcs);
 
	forall_edges(e,G1) 
	{ 

      
		if (negcycle_arcs.rank(e) != 0) // e in negcycle_arcs
		{
			ie=G1.inf(e);
			delta_sum+=ie.delta();
			lambda_sum+=ie.lambda();
			e2 = theEdge((in=G1.inf(G1.source(e))).id(), (in=G1.inf(G1.target(e))).id(), ie.lambda());
			el.append(e2); 
		}
	}
	if(lambda_sum==0){
		if(delta_sum==0) return 0; // circuit with zero latency and zero distance
		else{
			/*NoCriticalCycle nc(*this, el);
			throw nc;*/
			return -1;
		}
	}
	LEDA::rational h_dep(delta_sum,lambda_sum);
	/* restore surrounding loop dep */
	return h_dep;
}


LEDA::rational LOOP::CRITICAL_CYCLE() const
{
	int delta_sum = 0; 
	int lambda_sum = 0;
	INFO_EDGE ie;
	INFO_NODE in;
	int d;
	edge e1, e2, e;

	GRAPH<INFO_NODE, INFO_EDGE> G1=*this;
	
	if (PREPARE_CRITICAL_CYCLE(G1, *this) == MAXDOUBLE){
		/*NoCriticalCycle nc;	
		throw nc; // Graphe sans circuits
		*/
		return -1;
	}
	
	list<edge> negcycle_arcs;
	NEGATVE_CYCLE(G1,negcycle_arcs);
	
	forall_edges(e,G1) 
	{ 

      
		if (negcycle_arcs.rank(e) != 0) // e in negcycle_arcs
		{
			ie=G1.inf(e);
			lambda_sum+=ie.lambda();
			delta_sum+=ie.delta();			
		}
	}
	if(lambda_sum==0){
		if(delta_sum==0) return 0; // circuit with zero latency and zero distance
		else{
		/* // Circuit with null distance
			NoCriticalCycle nc;
			throw nc;*/
			return -1;
		}
	}
	LEDA::rational h_dep(delta_sum,lambda_sum);
	/* restore surrounding loop dep */
	return h_dep;
}




int LOOP::MaxLambda() const
{
	edge e;
	INFO_EDGE ie;
	int i=0, j;
	forall_edges(e, *this) 
	{
		if ((j=lambda(e))>i) 
			i=j;
	}
	return i;
}

int LOOP::MinLambda() const
{
	edge e;
	INFO_EDGE ie;
	int i=MAXINT, j;
	forall_edges(e, *this) 
	{
		if ((j=lambda(e))<i) 
			i=j;
	}
	return i;
}


edge LOOP::theEdge(int source, int dest, int dist) const
{
	node s= node_with_id(source), d = node_with_id(dest);
	edge e, ee;
	INFO_EDGE ie;
	bool found=false;
	int max_delta=-MAXINT;
  
	forall_out_edges(e, s){
		if ((target(e)==d)&&(dist==lambda(e))&&(max_delta<=(ie=inf(e)).sigma())){
			ee=e;
			max_delta=delta(e);
			found = true;
		}
	}
	if (found) return (ee);
	else {
		return (NULL);
	}
}


bool retime_ddg(LOOP &G, LEDA::node_array<int> &r){
	
	//GRAPH<node,edge> retimed_graph;
	node u, v, w, s;
	edge e;
	INFO_EDGE ie;
	node_matrix<int> all_pair_longuest_path;// for retiming computation
	edge_array<int> cost;// for retiming computation
	INSTRUCTIONS_TYPES it;
	
	// Now, we should comp�te the correct nonnegative edges distances by retiming
	// forall e=(u,v), the new distance is r(v)-r(u)+lambda(e) to be computed
	
	
	// Insert a top source node s)
	s=G.node_with_id(G.new_node(it.opcode_id()));
	// Insert an edge from s to each node in the retimed_graph
	forall_nodes(u, G){
		if(u==s) continue; // neglect s
		G.new_edge(s,u, ie);
	}
	
	all_pair_longuest_path.init(G);
	
	cost.init(G);
	// cost of an edge for retiming computation is equal to -lambda,
	// inverted to lambda to compute a longest path using shortest_path algorithms
	forall_edges(e, G){
		if (G.source(e)==s) cost[e]=0;
		else cost[e]= G.lambda(e);
	}
	
	if (! ALL_PAIRS_SHORTEST_PATHS_T(G, cost, all_pair_longuest_path))//since the costs are inverted,=> compute the the longest paths
		return false; // in case of error
	
	
	//now, r(u) = longestpath(s,u)
	r.init(G);
	forall_nodes(u, G){
		if (u==s) continue; //neglect s
		r[u]=-all_pair_longuest_path(s,u);
	}
	
	// Now, r has been computed, retime G
	forall_edges(e, G){
		u=G.source(e);
		v=G.target(e);
		if ((G.lambda(e) + r[v] - r[u])<0) {
		cerr<<"DDG internal retiming error: negative retimed distance. ";
			cerr<<"r("<<G.id(v)<<")="<<r[v]<<" - r("<<G.id(u)<<")="<<r[u]<<" + lambda= "<<G.lambda(e)<<endl;
			return false;
		}
		G.set_lambda(e, G.lambda(e) + r[v] - r[u]); //
	}
	G.del_node(s); //delete s the inserted sink
	return true;
}

bool retime_ddg(LOOP &G){
	
	//GRAPH<node,edge> retimed_graph;
	node u, v, w, s;
	edge e;
	INFO_EDGE ie;
	LEDA::node_array<int> r; //retiming array
	node_matrix<int> all_pair_longuest_path;// for retiming computation
	edge_array<int> cost;// for retiming computation
	INSTRUCTIONS_TYPES it;
	
	// Now, we should comp�te the correct nonnegative edges distances by retiming
	// forall e=(u,v), the new distance is r(v)-r(u)+lambda(e) to be computed
	
	
	// Insert a top source node s)
	s=G.node_with_id(G.new_node(it.opcode_id()));
	// Insert an edge from s to each node in the retimed_graph
	forall_nodes(u, G){
		if(u==s) continue; // neglect s
		G.new_edge(s,u, ie);
	}
	
	all_pair_longuest_path.init(G);
	
	cost.init(G);
	// cost of an edge for retiming computation is equal to -lambda,
	// inverted to lambda to compute a longest path using shortest_path algorithms
	forall_edges(e, G){
		if (G.source(e)==s) cost[e]=0;
		else cost[e]= G.lambda(e);
	}
	
	if (! ALL_PAIRS_SHORTEST_PATHS_T(G, cost, all_pair_longuest_path))//since the costs are inverted,> compute the the longest paths
		return false; // in case of error
	
	
	//now, r(u) = longestpath(s,u)
	r.init(G);
	forall_nodes(u, G){
		if (u==s) continue; //neglect s
		r[u]=-all_pair_longuest_path(s,u);
	}
	
	// Now, r has been computed, retime G
	forall_edges(e, G){
		u=G.source(e);
		v=G.target(e);
		if ((G.lambda(e) + r[v] - r[u])<0) {
		cerr<<"DDG internal retiming error: negative retimed distance. ";
			cerr<<"r("<<G.id(v)<<")="<<r[v]<<" - r("<<G.id(u)<<")="<<r[u]<<" + lambda= "<<G.lambda(e)<<endl;
			return false;
		}
		G.set_lambda(e, G.lambda(e) + r[v] - r[u]); //
	}
	G.del_node(s); //delete s the inserted sink
	return true;
}


bool apply_retiming(LOOP &G, const LEDA::node_array<int> r){
	
	//GRAPH<node,edge> retimed_graph;
	node u, v, w;
	edge e;
	
	
	
	forall_edges(e, G){
		u=G.source(e);
		v=G.target(e);
		if ((G.lambda(e) + r[v] - r[u])<0) {
		cerr<<"DDG internal retiming error: negative retimed distance. ";
			cerr<<"r("<<G.id(v)<<")="<<r[v]<<" - r("<<G.id(u)<<")="<<r[u]<<" + lambda= "<<G.lambda(e)<<endl;
			return false;
		}
		G.set_lambda(e, G.lambda(e) + r[v] - r[u]); //
	}
	return true;
}


//returns true if all circuits C have \lambda(C)>0. Returns false otherwise
// the returned edge e is the delinquent edge: it belongs to the nonpositive circuit

bool is_lexicographic_positive( LOOP &G,  edge &e){

/* A retiming r for a DDG must satisfy r(v)-r(u) + lambda(e) >= 0. Thus, r(v)-r(u)=>lambda(e)
 This is equivalent to compute a potential p=r. For this purpose, we create a dummy source s, connected to each node with a cost equal to 0, and we compute the longest path from s to every node. The longest path is equal to the potential, which is the retiming. 
 
 since LEDA 5.2 does not have a longest path method, we use a shortest path algorithm by inverting the signs of the costs.
*/

/* A DDG is lexicographic positive iff forall retiming r_e (forall e=(u,v) \in E)
  r_e(v)-r_e(u) + lambda(e) >= 1
  forall e'=(u',v') != e:  r_e(v')-r(u') + lambda(e') >= 0
  
  A retiming r_e is computed as a potential such that :
  - the edge e has a cost equal to 1-lambda(e)
  - the edge e' != e has a cost equal to -lambda(e')
  
  
*/

	bool lex_positive = true;
	
	node s; // this is is a dummy source connected to every 
	int s_id;
	node u;
	edge ee, eee;
	INFO_EDGE ie;
	// Bellman Ford shortest_path algorithm requires what follows
	edge_array<int> cost;
	node_array<int> dist;
	node_array<edge> pred;
	
	// insert the dummy nodes and edges
	s_id=G.new_node(G.get_ISA().pick_an_instruction_type_id());
	s=G.node_with_id(s_id);
	
	forall_nodes(u,G){
		if(u!=s){
			G.new_edge(s,u, ie);
		}
	}
	
	cost.init (G);
	dist.init(G);
	pred.init(G);
	
	forall_edges(ee, G){ // Compute a retiming r_e
		if (G.lambda(ee)>0) continue; // treat only nonpositive arcs
		cost[ee]=  G.lambda(ee)-1; // invert the cost in order to compute a longest path instead of a shortest path
		forall_edges(eee, G){ // Compute a retiming r_e
			if (eee!=ee)
			cost[eee]=  G.lambda(eee); // invert the cost in order to compute a longest path instead of a shortest path
		}
		lex_positive=lex_positive&& BELLMAN_FORD_T(G, s, cost, dist, pred);
		
		if(!lex_positive){
			G.del_node(s);
			e=ee;
			
			return false;
		}
	}
	
	G.del_node(s);
	return true;
}
	
	



//returns true if all circuits C have \lambda(C)>0. Returns false otherwise
// the returned edge list le is an example of a delinquent circuit C where \lambda(C)<=0
///@todo
bool is_lexicographic_positive( LOOP &G,  leda::list<edge> &el){
	//leda::rational mii=G.CRITICAL_CYCLE(el);
	//debug
// 	cout<<"MII="<<mii<<endl;
// 	if(mii==0) return false;
// 	else return true;
/* A retiming r for a DDG must satisfy r(v)-r(u) + lambda(e) >= 0. Thus, r(v)-r(u)=>lambda(e)
	This is equivalent to compute a potential p=r. For this purpose, we create a dummy source s, connected to each node with a cost equal to 0, and we compute the longest path from s to every node. The longest path is equal to the potential, which is the retiming. 
 
	since LEDA 5.2 does not have a longest path method, we use a shortest path algorithm by inverting the signs of the costs.
*/

/* A DDG is lexicographic positive iff forall retiming r_e (forall e=(u,v) \in E)
	r_e(v)-r_e(u) + lambda(e) >= 1
	forall e'=(u',v') != e:  r_e(v')-r_e(u') + lambda(e') >= 0
  
  A retiming r_e is computed as a potential such that :
	- the edge e has a cost equal to 1-lambda(e)
	- the edge e' != e has a cost equal to -lambda(e')
  
  
*/

	bool lex_positive = true;
	
	node s; // this is is a dummy source connected to every 
	int s_id;
	node u;
	edge ee, eee;
	INFO_EDGE ie;
	// Bellman Ford shortest_path algorithm requires what follows
	edge_array<int> cost;
	node_array<int> dist;
	node_array<edge> pred;
	
	// insert the dummy nodes and edges
	s_id=G.new_node(G.get_ISA().pick_an_instruction_type_id());
	s=G.node_with_id(s_id);
	
	forall_nodes(u,G){
		if(u!=s){
			G.new_edge(s,u, ie);
		}
	}
	
	cost.init (G);
	dist.init(G);
	pred.init(G);
	
	forall_edges(ee, G){ // Compute a retiming r_e
		if (G.lambda(ee)>0) continue; // treat only nonpositive arcs
		cost[ee]=  G.lambda(ee)-1; // invert the cost in order to compute a longest path instead of a shortest path
		forall_edges(eee, G){ // Compute a retiming r_e
			if (eee!=ee)
				cost[eee]=  G.lambda(eee); // invert the cost in order to compute a longest path instead of a shortest path
		}
		lex_positive=lex_positive&& BELLMAN_FORD_T(G, s, cost, dist, pred);
		
		if(!lex_positive){
			//construct the delinquent circuit C
			el.clear();
			el.push(ee);
			u=source(ee);
			ee=pred[u];
			while(ee!=nil && G.source(ee)!=u){
				el.push(ee);
				ee=pred[G.source(ee)];
			}
			//destroy the source node before exiting
			G.del_node(s);			
			return false;
		}
	}
	
	G.del_node(s);
	return true;
}
	
	


}


int DDG::LOOP::lambda(edge e) const
{
	INFO_EDGE ie=inf(e);
	return ie.lambda();
}


void DDG::LOOP::set_lambda(edge e, int l)
{
	INFO_EDGE ie=inf(e);
	ie.set_lambda(l);
	LEDA::GRAPH<INFO_NODE, INFO_EDGE>::assign(e, ie);
}




/*!
    \fn DDG::LOOP::check()
    \returns true if the LOOP object seems OK.
The current implementation always return true.
 */
bool DDG::LOOP::check()
{
    /// @todo implement me
    return true;
}

/*!
    \fn DDG::LOOP::check(int t)
    \returns true if the LOOP object seems OK for the given register type.
The current implementation always return true.
 */
bool DDG::LOOP::check(int t)
{
    /// @todo implement me
	return true;
}
