/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/
/*! \file loop_example.cpp
\brief A simple code example to show how to use the LOOP class
*/
#include "DDG.h" // the header file of the DDG library. To be included by the user.

#include <iostream>
#include <cstdlib>


using namespace std;
using namespace DDG; // the namespace of the DDG library.



int main(int argc, char *argv[])
{
	//int are always usefull
	int         i,j ;
	int c; // for getopt
	// some strings for file names
	ARCHITECTURE isa_arch;
	
	char        str[MAX_STRING];
	char        *ddg_filename=NULL, *arch_filename=NULL;
	LEDA::list<REGISTER_TYPES> T;
	REGISTER_TYPES rt;
	// some objects for graphs
	//declare a DDG of a loop (cyclic data dependence graph)
	DDG::LOOP loop, unrolled, loop_copy;
	ARCHITECTURE isa_example;

	LEDA::node n,u;
	LEDA::edge e;
	set<node> VR;
	set<edge> ER;
	DDG::INFO_EDGE ie; //information attached to a given edge

	INSTRUCTIONS_TYPES it; // information attached to an opcode
	LEDA::list<edge> el; // a list of edges
	
	while ((c = getopt (argc, argv, "i:a:")) != -1){
		switch(c){
			case 'a': arch_filename=optarg;
				break;
			case 'i': ddg_filename =optarg;
				break;
			case '?': 
			 	cout<<"usage:"<< argv[0] <<" -i ddg_filename.xml [-a isa_desc.xml] "<<endl;
			 	return EXIT_FAILURE;
		}
		
	}
	
	if (arch_filename!=NULL)  { 
		isa_example.read_from_xml(arch_filename);
		if(isa_example.check()){
			loop=LOOP(isa_example);	 //build an empty DDG loop with a user defined architcural (ISA) description
 		} else return EXIT_FAILURE;
		
	} 
	 //---- The user can read a DDG from an xml format defined by the DDG library
	if (ddg_filename != NULL){
		i=loop.read_from_xml(ddg_filename); // reads the DDG loop from an xml file
		if(i==-1) return EXIT_FAILURE;

	}
	else{
	    cout<<"usage:"<< argv[0] <<" -i ddg_filename.xml [-a isa_desc.xml] "<<endl;
	    return EXIT_FAILURE;
	}

	

	 
	 // The user can modify the DDG by many other methods
	// : new_node, new_edge, clear, del_node, del_edge, etc.
	 // The user can access any information about nodes, edges,
	// apply well known algorithms using the LEDA library
	 // See the LEDA manual : http://www.algorithmic-solutions.info/leda_manual
	 // See the DDG user manual inside this distribution
  	
	cout<< "DDG example program: " <<ddg_filename<<endl;
	cout<< "------------------------------" <<endl;
	// ------------------- The user can compute the critical path ---------------
	// Null circuits are accepted inside the DDG and detected
	cout<<"Critical Cycle  of "<< ddg_filename<<" = "<<loop.CRITICAL_CYCLE(el)<<endl;
	cout<<"Edges belonging the the critical cycle"<<endl;
	forall(e, el){
		ie=loop[e];
		cout<< loop.source_id(e)<< " -> "<< loop.target_id(e);
		cout<<" : "<< ie<<endl;
	}

	
	//------- The user can outputs the DDGs to a .xml format --------
	snprintf(str, MAX_STRING, "%s_temp.xml", ddg_filename);;
	loop.write_to_xml(str); // Outputs the DDG loop to a .xml file format

	i=loop.read_from_xml(str); // RE-reading the DDG loop from a .xml file erases the previous DDG loop.
	if(i==-1) return EXIT_FAILURE;
	
	// ------ The user can visualize the DDGs ---------------------
	snprintf(str, MAX_STRING, "%s.vcg",ddg_filename); // sets the .vcg filename for the ddg loop
	loop.write_to_vcg(str); //outputs the DDG loop to a vcg file to be visualized by the xvcg tool
	cout<<endl<<str<< " can be visualized using xvcg."<<endl;
	
	loop_copy.copy(loop); // create an exact copy of the loop
	snprintf(str, MAX_STRING, "%s_copy.vcg",ddg_filename);
	//exports the copy loop to a vcg file to be visualized by the xvcg tool
	loop_copy.write_to_vcg(str);
	
	// The user can access any kind of information 
	T=loop.T();
	forall(rt, T){
		cout <<"The set of values of type "<< rt.get_register_typename() <<" is :"<<endl;
		VR=loop.get_V_R(rt.get_register_type_id());
		forall(u, VR){
			it=loop.instruction_type(u);
			cout<< loop.id(u)<< " with opcode " << it.opcode()<< " ("<<loop.latency(u)<<","<<loop.delta_r(u)<<","<<loop.delta_w(u, rt.get_register_type_id())<<")"<<endl;
		}
	}
	//----- We can unroll the loop while preserving recurrent edges
	unroll(loop, unrolled, 3);
	snprintf(str, MAX_STRING, "%s_unrolled_3.vcg",ddg_filename); // sets the .vcg filename for the ddg loop
	unrolled.write_to_vcg(str); //outputs the DDG loop to a vcg file to be visualized by the xvcg tool
	cout<<endl<<str<< " can be visualized using xvcg."<<endl;
	forall (rt, T){	
		cout <<"The set of flow edges of the loop of type "<< rt.get_register_typename()<<" is :"<<endl;
		ER=unrolled.get_E_R(rt.get_register_type_id());
		forall(e, ER){
		ie=unrolled[e];
		cout<< unrolled.source_id(e)<<" -> "<< unrolled.target_id(e)<<" : "<< ie<<endl;
		}
	}

	// ------------- Hiding and restoring nodes -------------
	// look for two arbitrary nodes to hide
	forall (rt, T){	
		
		if (loop.get_V_R(rt.get_register_type_id()).size()>=2) {
		loop.hide_node(loop.get_V_R(rt.get_register_type_id()).choose()); // just to hide one arbitrary value node. Notice that all adjacent edges are hidden too
		loop.hide_node(loop.get_V_R(rt.get_register_type_id()).choose()); // hide another arbitrary one
		break;
		}
	}
	cout<<"Critical Cycle  of "<< ddg_filename<<" (after hiding two arbitrary values of type "<< rt.get_register_typename()<<" ) = "<<loop.CRITICAL_CYCLE(el)<<endl;
	cout<<"Edges belonging the the critical cycle"<<endl;
	forall(e, el){
		ie=loop[e];
		cout<< loop.source_id(e)<< " -> "<< loop.target_id(e);
		cout<<" : "<< ie<<endl;
	}
		
	
	loop.restore_all_nodes();
	loop.restore_all_edges();

	cout<<"Critical Cycle  of "<< ddg_filename<<" (after restoring all values and edges) = "<<loop.CRITICAL_CYCLE(el)<<endl;
	cout<<"Edges belonging the the critical cycle"<<endl;
	forall(e, el){
		ie=loop[e];
		cout<< loop.source_id(e)<< " -> "<< loop.target_id(e);
		cout<<" : "<< ie<<endl;
	}		

	
	//--- Many other software opportunities exist. See the reference manual.

	return EXIT_SUCCESS;
}

