/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/
 
/*! \file loop_unroll.cpp
\brief A simple code example to show how to unroll a loop
*/
#include "DDG.h" // the header file of the DDG library. To be included by the user.

#include <iostream>
#include <cstdlib>


using namespace std;
using namespace DDG; // the namespace of the DDG library.

char *getStringForArgument(char * s, int argc, char *argv[])
 {
	for (int i = 1 ; i<argc; i++)
	{
		if (strcmp(argv[i],s) == 0)
		{ 
			if (i+1<argc){
				return argv[i+1];}
				else
					fprintf(stderr,"Argument value missing for %s\n",s);
				return(NULL);
		}
	}
	return(NULL);
}


int main(int argc, char *argv[])
{
	// some strings for file names
	char        *ifi;
	char        cc[400], str[400];
	
	// some objects for graphs
	DDG::LOOP loop, unrolled_loop; //declare a DDG of a loop (cyclic data dependence graph)
	DDG::ARCHITECTURE isa_arch;
	
	ifi = getStringForArgument("-i",argc,argv);
	if (ifi==NULL)  { // no line commande arguments
		cout<<"loop_unroll reads a gl loop and outputs an unrolled loop"<<endl;
		cout<<"usage: loop_unroll -i inputDDG.gl -o outputDDG.gl -U #unroll_factor [-a arch_desc]"<<endl;
		return EXIT_FAILURE;
	}
	strcpy(cc,ifi);
	
	ifi = getStringForArgument("-a",argc,argv);
	//check wether the user specifies an ISA description or not
	if (ifi!=NULL)  {
		//build an empty DDG loop with a user defined ISA description
		isa_arch.read_from_arch(ifi);
		if(isa_arch.check()){
			loop=LOOP(isa_arch);
			unrolled_loop=LOOP(isa_arch);
		}
		else exit(1);
	}
	
	loop.read_from_gl(cc); // reads the DDG loop from a gl file

	ifi = getStringForArgument("-o",argc,argv);
	if (ifi==NULL)  { // no line commande arguments
		cout<<"loop_unroll reads a gl loop and outputs an unrolled loop"<<endl;
		cout<<"usage: loop_unroll -i inputDDG.gl -o outputDDG.leda -U #unroll_factor [-a arch_desc]"<<endl;
		return EXIT_FAILURE;
	}

	strcpy(cc,ifi);
	
	ifi = getStringForArgument("-U",argc,argv);
	if (ifi==NULL)  { // no line commande arguments
		cout<<"loop_unroll reads a gl loop and outputs an unrolled loop"<<endl;
		cout<<"usage: loop_unroll -i inputDDG.gl -o outputDDG.leda -U #unroll_factor [-a arch_desc]"<<endl;
		return EXIT_FAILURE;
	}

	unroll(loop, unrolled_loop, atoi(ifi));
	unrolled_loop.write_to_gl(cc);
	return EXIT_SUCCESS;
}

