/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/

/*! \file register_types.cpp
\brief Methods implementations for REGISTER_TYPES class.
A REGISTER_TYPES object can be accessed and modified using the methods implemented in the current file.
*/
#include "register_types.h"
using namespace std;

namespace DDG {
// 
REGISTER_TYPES::REGISTER_TYPES(): list_registers()
{
	nb_registers=-1;
	register_type_id=-1;
}

REGISTER_TYPES::REGISTER_TYPES(std::string regtypename): list_registers()
{
	nb_registers=-1;
	register_type_id=-1;
	register_typename=regtypename;
}

REGISTER_TYPES::REGISTER_TYPES(std::string regtypename, int nbreg): list_registers()
{
	nb_registers=nbreg;
	register_type_id=-1;
	register_typename=regtypename;
}

REGISTER_TYPES::REGISTER_TYPES(int nbreg): list_registers()
{
	nb_registers=nbreg;
	register_type_id=-1;
}

REGISTER_TYPES::~REGISTER_TYPES()
{
}




/*!
    \fn DDG::REGISTER_TYPES::clear()
    Makes it empty.
 */
void REGISTER_TYPES::clear()
{
    	list_registers.clear();
	nb_registers=0;
	register_type_id=-1;
}


/*!
    \fn DDG::REGISTER_TYPES::check()
    Simple check method.
\returns true if the REGISTER_TYPES seems OK. False otherwise.

 */
bool REGISTER_TYPES::check() const
{
    //if(nb_registers<=0) return false;
    if (register_typename.empty()) {
    	cerr << "DDG warning : registertype check find an empty name.\n" ;
    	return false;
    }
   
    if (register_type_id<0) {
    	cerr << "DDG warning : registertype check for "<<register_typename<<" : negative id not allowed.\n" ;
    	return false;
    }
    
    //if(list_registers.size()<0 && nb_registers!=list_registers.size()) return false;
    
    return true;
}


/*!
    \fn DDG::REGISTER_TYPES::remove_register_name(std::string reg_name)
    Removes the given architectural register from the list of registers.
 */
void REGISTER_TYPES::remove_register_name(std::string reg_name)
{
    list_registers.remove(reg_name);
    
    
}

 /*!
        \fn DDG::REGISTER_TYPES::add_register_name(std::string reg_name)
       Add the given architectural register to the list of registers. This method allows to construct a list of register names of a given register type.
 */
void REGISTER_TYPES::add_register_name(std::string reg_name)
{
	list_registers.push(reg_name);
}

ostream& operator<<(ostream &os,  REGISTER_TYPES x){
os<<x.get_register_typename()<< " " << x.get_register_type_id() <<" "<< x.get_nb_registers();
}

istream& operator>>(istream &is, REGISTER_TYPES & x){
std::string name;
int id, nbreg;
is>> name;
is>> id>>nbreg;
x.set_register_typename(name);
x.set_register_type_id(id);
x.set_nb_registers(nbreg);
}


/*!
	\fn operator==(const REGISTER_TYPES t1, const REGISTER_TYPES t2)
	\returns true iff the two register types t1 and t2 have the same id or have the same name.
 */
bool operator==(const REGISTER_TYPES t1, const REGISTER_TYPES t2) {
	return ((t1.get_register_type_id()== t2.get_register_type_id())
			|| (t1.get_register_typename()==t2.get_register_typename())
	       );
}

}
