/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/

#ifndef __register_types_H
#define __register_types_H

/*! \file register_types.h
\brief The header file of register types class.
This file contains the method declaration to construct an object of register type.
*/


#endif
#ifndef DDGREGISTER_TYPES_H
#define DDGREGISTER_TYPES_H

#include <LEDA/core/list.h>
#include <iostream>

namespace DDG {

/**
	@author Touati Sid <Sid-nospam.Touati-nospam@inria.fr>
*/
class REGISTER_TYPES{
private:
    int nb_registers;
    int register_type_id;
    LEDA::list<std::string> list_registers;
    std::string register_typename;
public:
    	REGISTER_TYPES();
	REGISTER_TYPES(int nbreg);
	REGISTER_TYPES(std::string regtypename);
	REGISTER_TYPES(std::string regtypename, int nbreg);
    	~REGISTER_TYPES();
	
	void set_nb_registers ( int theValue )
	{
		nb_registers = theValue;
	}
	

	int get_nb_registers() const
	{
		return nb_registers;
	}
	
  LEDA::list<std::string> get_list_registers(){
    return list_registers;
  }
	
	void set_register_typename( std::string theValue )
	{
		register_typename = theValue;
	}
	

	std::string get_register_typename() const
	{
		return register_typename;
	}
	const char * get_str_register_typename() const{  return (register_typename.c_str());}
    void clear() ;
    bool check() const;
    void remove_register_name(std::string reg_name);

   
    void add_register_name(std::string reg_name);

	void set_register_type_id ( int theValue )
	{
		register_type_id = theValue;
	}
	

	int get_register_type_id() const
	{
		return register_type_id;
	}
	
    	friend std::ostream& operator<<(std::ostream &os,  REGISTER_TYPES x);
	friend std::istream& operator>>(std::istream &is, REGISTER_TYPES & x);	

};
bool operator==(const REGISTER_TYPES t1, const REGISTER_TYPES t2);
}

#endif
