/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/


/*! \file xml2vcg.cpp
\brief A simple code example to show how to export a DDG file from xml format to vcg format to be visualized with xvcg
*/
#include "DDG.h" // the header file of the DDG library. To be included by the user.

#include <iostream>
#include <cstdlib>


using namespace std;
using namespace DDG; // the namespace of the DDG library.



int main(int argc, char *argv[])
{
	// some strings for file names
	char        *ifi;
	char        str[400];
	char        *ddg_filename=NULL, *arch_filename=NULL, *output_filename=NULL;
	
	// some objects for graphs
	DDG::LOOP loop; //declare a DDG of a loop (cyclic data dependence graph)
	ARCHITECTURE isa_example;
	int c;
	
	while ((c = getopt (argc, argv, "i:a:o:")) != -1){
		switch(c){
			case 'a': arch_filename=optarg;
				break;
			case 'i': ddg_filename =optarg;
				break;
			case 'o': output_filename=optarg;
				break;
			case '?': 
			 	cout<<"usage:"<< argv[0] <<" -i ddg_filename.xml o vcg_filename.vcg [-a isa_desc.xml] "<<endl;
			 	return EXIT_FAILURE;
		}
		
	}
	
	if (arch_filename!=NULL)  { 
		isa_example.read_from_xml(arch_filename);
		if(isa_example.check()){
			loop=LOOP(isa_example);	 //build an empty DDG loop with a user defined architcural (ISA) description
 		} else return EXIT_FAILURE;
		
	} 
	 //---- The user can read a DDG from an xml format defined by the DDG library
	if (ddg_filename != NULL){
		c=loop.read_from_xml(ddg_filename); // reads the DDG loop from an xml file
		if(c==-1) return EXIT_FAILURE;

	}
	else{
	    cout<<"usage:"<< argv[0] <<" -i ddg_filename.xml -o vcg_filename.vcg [-a isa_desc.xml] "<<endl;
	    return EXIT_FAILURE;
	}

	
	if (output_filename == NULL){
		cout<<"usage:"<< argv[0] <<" -i ddg_filename.xml -o vcg_filename.vcg [-a isa_desc.xml] "<<endl;
	    return EXIT_FAILURE;
	}
	
	
	
	c=loop.read_from_xml(ddg_filename); // reads the DDG loop from a xml file
	if(c==-1) return EXIT_FAILURE;
	
	//outputs the DDG loop to a vcg file to be visualized by the xvcg tool
	loop.write_to_vcg(output_filename);
	return EXIT_SUCCESS;
}

